#
# Copyright (c) 2013+ Anton Tyurin <noxiouz@yandex.ru>
# Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
# Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
# This file is part of Cocaine-tools.
#
# Cocaine is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Cocaine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import logging
import os
import re
import shutil
import tarfile
import tempfile

import msgpack
from tornado import gen

from cocaine.decorators import coroutine
# from cocaine.exceptions import LocatorResolveError
from cocaine.exceptions import ServiceError
from cocaine.tools import actions, log
from cocaine.tools.actions import common, readArchive, CocaineConfigReader, docker
from cocaine.tools.actions.common import NodeInfo
from cocaine.tools.error import ToolsError
from cocaine.tools.installer import PythonModuleInstaller, ModuleInstallError, _locateFile
from cocaine.tools.printer import printer
from cocaine.tools.repository import GitRepositoryDownloader, RepositoryDownloadError
from cocaine.tools.tags import APPS_TAGS

__author__ = 'Evgeny Safronov <division494@gmail.com>'

WRONG_APPLICATION_NAME = 'Application "{0}" is not valid application'

venvFactory = {
    'None': None,
    'P': PythonModuleInstaller,
    'R': None,
    'J': None
}


# class Specific(actions.Specific):
#     def __init__(self, storage, name):
#         super(Specific, self).__init__(storage, 'application', name)


class List(actions.List):
    def __init__(self, storage):
        super(List, self).__init__('manifests', APPS_TAGS, storage)


class View(actions.View):
    def __init__(self, storage, name):
        super(View, self).__init__(storage, 'application', name, 'manifests')


class Upload(actions.Storage):
    """
    Storage action class that tries to upload application into storage asynchronously
    """

    def __init__(self, storage, name, manifest, package=None, manifest_only=False):
        super(Upload, self).__init__(storage)
        self.name = name
        self.manifest = manifest
        if manifest_only:
            self.package = None
        else:
            self.package = package
            if not self.package:
                raise ValueError('Please specify package of the app')

        if not self.name:
            raise ValueError('Please specify name of the app')
        if not self.manifest:
            raise ValueError('Please specify manifest of the app')

    @coroutine
    def execute(self):
        with printer('Loading manifest'):
            manifest = CocaineConfigReader.load(self.manifest)

        #  Not only a manifest is being uploaded,
        #  self.package could be None if manifest_only=True
        if self.package is not None:
            with printer('Reading package "%s"', self.package):
                package = msgpack.dumps(readArchive(self.package))

            with printer('Uploading application "%s"', self.name):
                channel = yield self.storage.write('apps', self.name, package, APPS_TAGS)
                yield channel.rx.get()

        with printer('Uploading manifest'):
            channel = yield self.storage.write('manifests', self.name, manifest, APPS_TAGS)
            yield channel.rx.get()

        raise gen.Return("Uploaded successfully")


class Remove(actions.Storage):
    """
    Storage action class that removes application 'name' from storage
    """

    def __init__(self, storage, name):
        super(Remove, self).__init__(storage)
        self.name = name
        if not self.name:
            raise ValueError('Empty application name')

    @coroutine
    def execute(self):
        with printer('Removing "%s"', self.name):
            apps = yield List(self.storage).execute()
            if self.name not in apps:
                raise ToolsError('application "{0}" does not exist'.format(self.name))
            channel = yield self.storage.remove('manifests', self.name)
            yield channel.rx.get()
            try:
                channel = yield self.storage.remove('apps', self.name)
                yield channel.rx.get()
            except ServiceError:
                log.info('Unable to delete an application source from storage. '
                         'It\'s okay, if the application is a Docker image')

        raise gen.Return("Removed successfully")


class Start(common.Node):
    def __init__(self, node, name, profile):
        super(Start, self).__init__(node)
        self.name = name
        self.profile = profile
        if not self.name:
            raise ValueError('Please specify application name')
        if not self.profile:
            raise ValueError('Please specify profile name')

    @coroutine
    def execute(self):
        channel = yield self.node.start_app(self.name, self.profile)
        yield channel.rx.get()
        raise gen.Return("application `%s` has been started with profile `%s`" % (self.name,
                                                                                  self.profile))


class Stop(common.Node):
    def __init__(self, node, name):
        super(Stop, self).__init__(node)
        self.name = name
        if not self.name:
            raise ValueError('Please specify application name')

    @coroutine
    def execute(self):
        channel = yield self.node.pause_app(self.name)
        yield channel.rx.get()
        raise gen.Return("application `%s` has been stopped" % self.name)


class Restart(common.Node):
    def __init__(self, node, locator, name, profile):
        super(Restart, self).__init__(node)
        self.locator = locator
        self.name = name
        self.profile = profile
        if not self.name:
            raise ValueError('Please specify application name')

    @coroutine
    def execute(self):
        try:
            if self.profile:
                profile = self.profile
            else:
                info = yield NodeInfo(self.node, self.locator, self.name).execute()
                app_info = info['apps'][self.name]
                if not isinstance(app_info, dict):
                    raise ToolsError('Unable to determine a profile name from info: {0}'.format(app_info))
                profile = app_info['profile']['name']
            try:
                yield Stop(self.node, name=self.name).execute()
            except ServiceError as err:  # application is not running
                pass
            yield Start(self.node, name=self.name, profile=profile).execute()
        except KeyError:
            raise ToolsError('Application "{0}" is not running and profile not specified'.format(self.name))
        except ServiceError as err:
            raise ToolsError('Unknown error - {0}'.format(err))

        raise gen.Return("application `%s` has been restarted with profile `%s`" % (self.name, profile))


class Check(common.Node):
    def __init__(self, node, name):
        super(Check, self).__init__(node)
        self.name = name
        if not self.name:
            raise ValueError('Please specify application name')

    @coroutine
    def execute(self):
        try:
            channel = yield self.node.list()
            apps = yield channel.rx.get()
            if self.name not in apps:
                raise ToolsError('not running')
        except ServiceError:
            raise ToolsError('not running')

        try:
            channel = yield self.node.info(self.name)
            info = yield channel.rx.get()
            log.info(info['state'])
        except ServiceError:
            raise ToolsError('stopped')
        raise gen.Return(info)


class DockerUpload(actions.Storage):
    def __init__(self, storage, path, name, manifest, address, registry='', on_read=None):
        super(DockerUpload, self).__init__(storage)
        self.path = path or os.path.curdir
        self.name = name or os.path.basename(os.path.abspath(self.path))
        if registry:
            self.fullname = '{0}/{1}'.format(registry, self.name)
        else:
            self.fullname = self.name

        self.manifest = manifest

        self.client = docker.Client(address)

        log.debug('checking Dockerfile')
        if not address:
            raise ValueError('Docker address is not specified')

        if on_read is not None:
            if not callable(on_read):
                raise ValueError("on_read must ne callable")
            self._on_read = on_read

        self._last_message = ''

    @coroutine
    def execute(self):
        log.debug('application name will be: %s', self.fullname)

        if self.manifest:
            manifestPath = self.manifest
        else:
            try:
                manifestPath = _locateFile(self.path, 'manifest.json')
            except IOError:
                log.error("unable to locate manifest.json")
                raise ToolsError("unable to locate manifest.json")

        with printer('Loading manifest'):
            manifest = CocaineConfigReader.load(manifestPath)

        with printer('Uploading manifest'):
            channel = yield self.storage.write('manifests', self.name, manifest, APPS_TAGS)
            yield channel.rx.get()

        try:
            response = yield self.client.build(self.path, tag=self.fullname, streaming=self._on_read)
            if response.code != 200:
                raise ToolsError('building failed with error code {0} {1}'.format(response.code,
                                                                                  response.body))
            response = yield self.client.push(self.fullname, auth={}, streaming=self._on_read)
            if response.code != 200:
                raise ToolsError('pushing failed with error code {0} {1}'.format(response.code,
                                                                                 response.body))
        except Exception as err:
            log.error("Error occurred. %s Erase manifest", err)
            channel = yield self.storage.remove('manifests', self.name)
            yield channel.rx.get()
            raise err

    def _on_read(self, value):
        if self._last_message != value:
            self._last_message = value
            print(value)


class DockerImport(actions.Storage):
    def __init__(self, storage, path, name, manifest, address, container, registry='', on_read=None):

        print "__init", storage, path, name, manifest, address, container

        super(DockerImport, self).__init__(storage)
        self.path = path or os.path.curdir
        self.name = name or os.path.basename(os.path.abspath(self.path))
        self.container_url = container
        if registry:
            self.fullname = '{0}/{1}'.format(registry, self.name)
        else:
            self.fullname = self.name

        self.manifest = manifest

        self.client = docker.Client(address)

        log.debug('checking Dockerfile')
        if not address:
            raise ValueError('Docker address is not specified')

        if on_read is not None:
            if not callable(on_read):
                raise ValueError("on_read must ne callable")
            self._on_read = on_read

        self._last_message = ''

    @coroutine
    def execute(self):
        log.debug('application name will be: %s', self.fullname)

        if self.manifest:
            manifestPath = self.manifest
        else:
            try:
                manifestPath = _locateFile(self.path, 'manifest.json')
            except IOError:
                log.error("unable to locate manifest.json")
                raise ToolsError("unable to locate manifest.json")

        with printer('Loading manifest'):
            manifest = CocaineConfigReader.load(manifestPath)

        with printer('Uploading manifest'):
            yield self.storage.write('manifests', self.name, manifest, APPS_TAGS)

        try:
            response = yield self.client.pull(self.container_url, {}, streaming=self._on_read)
            if response.code != 200:
                raise ToolsError('building failed with error code {0} {1}'.format(response.code,
                                                                                  response.body))

            response = yield self.client.tag(self.container_url, {}, self.fullname, streaming=self._on_read)
            if response.code != 200 and response.code != 201:
                raise ToolsError('building failed with error code {0} {1}'.format(response.code,
                                                                                  response.body))

            response = yield self.client.push(self.fullname, {}, streaming=self._on_read)
            if response.code != 200:
                raise ToolsError('pushing failed with error code {0} {1}'.format(response.code,
                                                                                 response.body))
        except Exception as err:
            log.error("Error occurred. Erase manifest")
            yield self.storage.remove('manifests', self.name)
            raise err

    def _on_read(self, value):
        if self._last_message != value:
            self._last_message = value
            print(value)


class LocalUpload(actions.Storage):
    def __init__(self, storage, path, name, manifest):
        super(LocalUpload, self).__init__(storage)
        self.path = path or os.path.curdir
        self.name = name
        self.manifest = manifest
        self.virtualEnvironmentType = 'None'
        if not self.name:
            self.name = os.path.basename(os.path.abspath(self.path))
        if not self.name:
            raise ValueError(WRONG_APPLICATION_NAME.format(self.name))

    @coroutine
    def execute(self):
        try:
            repositoryPath = self._createRepository()
            if self.manifest:
                manifestPath = self.manifest
            else:
                manifestPath = _locateFile(self.path, 'manifest.json')
            Installer = venvFactory[self.virtualEnvironmentType]
            if Installer:
                yield self._createVirtualEnvironment(repositoryPath, manifestPath, Installer)
                manifestPath = os.path.join(repositoryPath, 'manifest.json')
            else:
                pass

            packagePath = self._createPackage(repositoryPath)
            yield Upload(self.storage, **{
                'name': self.name,
                'manifest': manifestPath,
                'package': packagePath
            }).execute()
        except (RepositoryDownloadError, ModuleInstallError) as err:
            log.error(err)

    def _createRepository(self):
        with printer('Creating temporary directory') as p:
            repositoryPath = tempfile.mkdtemp()
            p(repositoryPath)

        with printer('Copying "%s" to "%s"', self.path, repositoryPath):
            repositoryPath = os.path.join(repositoryPath, 'repo')
            log.debug('Repository temporary path - "{0}"'.format(repositoryPath))
            shutil.copytree(self.path, repositoryPath)
            return repositoryPath

    @coroutine
    def _createVirtualEnvironment(self, repositoryPath, manifestPath, Installer):
        log.debug('Creating virtual environment "{0}"...'.format(self.virtualEnvironmentType))
        stream = None
        for handler in log.handlers:
            if isinstance(handler, logging.StreamHandler) and hasattr(handler, 'fileno'):
                stream = handler.stream
                break
        installer = Installer(path=repositoryPath, outputPath=repositoryPath, manifestPath=manifestPath, stream=stream)
        installer.install()

    def _createPackage(self, repositoryPath):
        with printer('Creating package'):
            packagePath = os.path.join(repositoryPath, 'package.tar.gz')
            tar = tarfile.open(packagePath, mode='w:gz')
            tar.add(repositoryPath, arcname='')
            tar.close()
            return packagePath


class UploadRemote(actions.Storage):
    def __init__(self, storage, path, name):
        super(UploadRemote, self).__init__(storage)
        self.url = path
        self.name = name
        if not self.url:
            raise ValueError('Please specify repository URL')
        if not self.name:
            rx = re.compile(r'^.*/(?P<name>.*?)(\..*)?$')
            match = rx.match(self.url)
            self.name = match.group('name')

    @coroutine
    def execute(self):
        repositoryPath = tempfile.mkdtemp()
        manifestPath = os.path.join(repositoryPath, 'manifest-start.json')
        packagePath = os.path.join(repositoryPath, 'package.tar.gz')
        self.repositoryDownloader = GitRepositoryDownloader()
        self.moduleInstaller = PythonModuleInstaller(repositoryPath, manifestPath)
        print('Repository path: {0}'.format(repositoryPath))
        try:
            yield self.cloneRepository(repositoryPath)
            yield self.installRepository()
            yield self.createPackage(repositoryPath, packagePath)
            yield Upload(self.storage, **{
                'name': self.name,
                'manifest': manifestPath,
                'package': packagePath
            }).execute()
        except (RepositoryDownloadError, ModuleInstallError) as err:
            print(err)

    @coroutine
    def cloneRepository(self, repositoryPath):
        self.repositoryDownloader.download(self.url, repositoryPath)

    @coroutine
    def installRepository(self):
        self.moduleInstaller.install()

    @coroutine
    def createPackage(self, repositoryPath, packagePath):
        tar = tarfile.open(packagePath, mode='w:gz')
        tar.add(repositoryPath, arcname='')



#
# Copyright (c) 2013+ Anton Tyurin <noxiouz@yandex.ru>
# Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
# Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
# This file is part of Cocaine-tools.
#
# Cocaine is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Cocaine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

__version__ = "0.12.13.6"



#
#    Copyright (c) 2013+ Anton Tyurin <noxiouz@yandex.ru>
#    Copyright (c) 2011-2013 Other contributors as noted in the AUTHORS file.
#
#    This file is part of Cocaine.
#
#    Cocaine is free software; you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    Cocaine is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from tornado import template

dockerchef = template.Template("""
FROM {{ basecontainer }}
RUN mkdir -p /tmp/chef
ADD ./{{ cookbooks }} /tmp/chef/
ADD ./solo.rb /tmp/chef/
ADD ./solo.json /tmp/chef/
RUN ls -l /tmp/chef/
# install chef-solo
RUN apt-get install curl -y
RUN curl -L https://www.opscode.com/chef/install.sh | bash
RUN chef-solo -c /tmp/chef/solo.rb -j /tmp/chef/solo.json
""")

dockerpuppet = template.Template("""
FROM {{ basecontainer }}
RUN mkdir -p /tmp/puppet

ADD ./puppet /tmp/puppet

RUN ls -l /tmp/puppet

# install puppet
RUN apt-get install puppet -y

# Copy modules
RUN if [ -d /tmp/puppet/modules ]; then \
        echo "Copy Puppet modules into /etc/puppet/modules/";\
        cp -Rv /tmp/puppet/modules/* /etc/puppet/modules/ ;\
    fi

# Apply manifest
RUN echo "Apply cocaine.pp manifest"
RUN puppet apply /tmp/puppet/cocaine.pp --modulepath=/etc/puppet/modules
""")



import logging


class ContextAdapter(logging.LoggerAdapter):
    def process(self, msg, kwargs):
        kwargs.setdefault("extra", {}).update(self.extra)
        return msg, kwargs


class NullLogger(object):
    def __call__(self, *args, **kwargs):
        return self

    def __getattribute__(self, name):
        return self

    def __setattr__(self, name, value):
        pass

    def __delattr__(self, name):
        pass


NULLLOGGER = NullLogger()



from cocaine.decorators import coroutine


class SecurePlugin(object):
    def __init__(self, repo):
        pass

    def ty(self):
        raise NotImplementedError

    @coroutine
    def fetch_token(self):
        raise NotImplementedError



#
# Copyright (c) 2013+ Anton Tyurin <noxiouz@yandex.ru>
# Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
# Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
# This file is part of Cocaine-tools.
#
# Cocaine is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Cocaine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from collections import Iterable

try:
    import simplejson as json
    from simplejson import JSONDecodeError
except ImportError:  # pragma: no cover
    import json
    JSONDecodeError = ValueError


class JSONUnpacker(Iterable):
    def __init__(self):
        self.buff = ""

    def feed(self, data):
        self.buff += data

    def __iter__(self):
        return self

    def next(self):
        js = json.JSONDecoder()
        try:
            res, index = js.raw_decode(self.buff)
            self.buff = self.buff[index:].lstrip('\r\n')
            return res
        except JSONDecodeError:
            raise StopIteration



#!/usr/bin/env python
#
# Copyright (c) 2013+ Anton Tyurin <noxiouz@yandex.ru>
# Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
# Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
# This file is part of Cocaine-tools.
#
# Cocaine is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Cocaine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import print_function

try:
    import httplib
except ImportError:
    import http.client as httplib  # pylint: disable=F0401

import collections
import datetime
import errno
import functools
import hashlib
import logging
import os
import random
import socket
import sys
import time

import msgpack
import tornado
from tornado import gen
from tornado import httputil
from tornado import process
from tornado.httpserver import HTTPServer
from tornado.iostream import StreamClosedError
from tornado.netutil import bind_sockets, bind_unix_socket
from tornado.util import import_object

try:
    from cocaine.logger import LoggerWithExtraInRecord
    logging.setLoggerClass(LoggerWithExtraInRecord)
except ImportError:
    print("current version of python framework does not provide LoggerWithExtraInRecord")

from cocaine.logger import CocaineHandler
from cocaine.logger import Logger
from cocaine.services import Service
from cocaine.services import Locator
from cocaine.exceptions import ServiceError
from cocaine.exceptions import DisconnectionError
from cocaine.services import EmptyResponse
from cocaine.detail.trace import Trace

from cocaine.proxy.helpers import Endpoints
from cocaine.proxy.helpers import extract_app_and_event
from cocaine.proxy.helpers import fill_response_in
from cocaine.proxy.helpers import write_chunked
from cocaine.proxy.helpers import finalize_chunked_response
from cocaine.proxy.helpers import header_to_seed
from cocaine.proxy.helpers import load_srw_config
from cocaine.proxy.helpers import pack_httprequest
from cocaine.proxy.helpers import parse_locators_endpoints
from cocaine.proxy.helpers import ProxyInvalidRequest
from cocaine.proxy.helpers import upper_bound
from cocaine.proxy.logutils import ContextAdapter
from cocaine.proxy.logutils import NULLLOGGER
from cocaine.proxy.plugin import IPlugin
from cocaine.proxy.plugin import PluginNoSuchApplication
from cocaine.proxy.plugin import PluginApplicationError
from cocaine.proxy.utilserver import UtilServer


try:
    from cocaine.tools.version import __version__ as tools_version
except ImportError:
    tools_version = "<undefinded>"


DEFAULT_SERVICE_CACHE_COUNT = 5
DEFAULT_REFRESH_PERIOD = 120
DEFAULT_TIMEOUT = 30
DEFAULT_TRACING_CHANCE = 5.  # %

_DEFAULT_BACKLOG = 128

# sec Time to wait for the response chunk from locator
RESOLVE_TIMEOUT = 5

# cocaine system category, I hope it will never change
SYSTEMCATEGORY = (0xff, 0xc)
EAPPSTOPPED = errno.EPIPE

LOCATORCATEGORY = (0xff, 0xa)
ESERVICENOTAVAILABLE = 1

OVERSEERCATEGORY = (0xff, 0x52ff)
EQUEUEISFULL = 1

# no such application
# we are mature enough to have our own status code
# but nginx proxy_next_upstream does NOT support custom codes
NO_SUCH_APP = httplib.SERVICE_UNAVAILABLE

X_COCAINE_HTTP_PROTO_VERSION = "X-Cocaine-HTTP-Proto-Version"


class BodyProcessor(object):

    def __init__(self, request, name, code, headers):
        self.request = request
        self.name = name
        self.code = code
        self.headers = headers
        self.messages = []

    def swallow(self, part):
        raise NotImplementedError

    def finish(self):
        raise NotImplementedError

    @staticmethod
    def make_processor(content_length, request, name, code, headers):
        if content_length is None and request.version != 'HTTP/1.0':
            return ChunkedBodyProcessor(request, name, code, headers)
        else:
            return CachedBodyProcessor(request, name, code, headers)


class ChunkedBodyProcessor(BodyProcessor):

    def __init__(self, request, name, code, headers):
        super(ChunkedBodyProcessor, self).__init__(
            request, name, code, headers)

        self.headers.add('Transfer-Encoding', 'chunked')
        self.headers['X-Cocaine-Application'] = self.name

        fill_response_in(self.request, self.code,
                         httplib.responses.get(self.code, httplib.OK),
                         ''.join(self.messages), self.headers, chunked=True)

    def swallow(self, part):
        write_chunked(self.request, part)

    def finish(self):
        finalize_chunked_response(
            self.request, self.code,
            httplib.responses.get(self.code, httplib.OK))


class CachedBodyProcessor(BodyProcessor):

    def swallow(self, part):
        self.messages.append(part)

    def finish(self):
        self.headers['X-Cocaine-Application'] = self.name

        fill_response_in(
            self.request, self.code,
            httplib.responses.get(self.code, httplib.OK),
            ''.join(self.messages), self.headers)


def proxy_error_headers(name=None):
    headers = {}
    if name is not None:
        headers['X-Cocaine-Application'] = name
    headers["X-Error-Generated-By"] = "Cocaine-Tornado-Proxy"
    return httputil.HTTPHeaders(headers)


def generate_request_id(request):
    data = "%d:%f" % (id(request), time.time())
    return hashlib.md5(data).hexdigest()


def get_request_id(request_id_header, request, force=False):
    return request.headers.get(request_id_header) or\
        (generate_request_id(request) if force else None)


def context(func):
    @gen.coroutine
    def wrapper(self, request):
        self.requests_in_progress += 1
        self.requests_total += 1
        traceid = None
        try:
            generated_traceid = self.get_request_id(request)
            if generated_traceid is not None:
                # assume we have hexdigest form of number
                # get only 16 digits
                traceid = generated_traceid[:16]
                adaptor = ContextAdapter(self.access_log, {"trace_id": traceid})
                request.logger = adaptor
                # verify user input: request_header must be valid hexdigest
                try:
                    int(traceid, 16)
                except ValueError:
                    fill_response_in(request, httplib.INTERNAL_SERVER_ERROR,
                                     httplib.responses[httplib.INTERNAL_SERVER_ERROR],
                                     "Request-Id `%s` is not a hexdigest" % traceid,
                                     proxy_error_headers())
                    return
            else:
                request.logger = NULLLOGGER  # pylint: disable=R0204
            request.traceid = traceid
            request.tracebit = True
            request.logger.info("start request: %s %s %s", request.host, request.remote_ip, request.uri)
            yield func(self, request)
        finally:
            self.requests_in_progress -= 1
    return wrapper


def gen_uid():
    return "proxy:%s_%d_%f" % (socket.gethostname(), os.getpid(), time.time())


def scan_for_updates(current, new):
    # add removed groups and new groups to updated
    # mark routing group as updated if its current ring is not
    # the same as new
    updated = [k for k, v in new.iteritems() if v != current.pop(k, None)]
    updated.extend(current.keys())
    return updated


def drop_app_from_cache(cache, app, name):
    # cache is defaultdict(list), so we can NOT rely on KeyError
    apps = cache.get(name)
    if apps is not None and app in apps:
        # remove app from cache
        apps.remove(app)
        # if there is no such apps in cache - remove key from dict
        # to avoid memory leak
        if len(apps) == 0:
            cache.pop(name)


def load_plugin(name, proxy, config):
    klass = import_object(name)
    if not issubclass(klass, IPlugin):
        raise Exception("%s is not a subclass of %s" % (klass.__name__, IPlugin.__name__))
    return klass(proxy, config)


class CocaineProxy(object):
    def __init__(self, locators=("localhost:10053",),
                 cache=DEFAULT_SERVICE_CACHE_COUNT,
                 request_id_header="", sticky_header="X-Cocaine-Sticky",
                 forcegen_request_header=False,
                 default_tracing_chance=DEFAULT_TRACING_CHANCE,
                 configuration_service="unicorn",
                 tracing_conf_path="/zipkin_sampling",
                 timeouts_conf_path="/proxy_apps_timeouts",
                 srw_config=None,
                 allow_json_rpc=True,
                 ioloop=None, **config):
        # stats
        self.requests_in_progress = 0
        self.requests_disconnections = 0
        self.requests_total = 0

        self.io_loop = ioloop or tornado.ioloop.IOLoop.current()
        self.service_cache_count = cache
        self.spool_size = int(self.service_cache_count * 1.5)
        self.refresh_period = config.get("refresh_timeout", DEFAULT_REFRESH_PERIOD)
        self.locator_endpoints = [parse_locators_endpoints(i) for i in locators]
        # it's initialized after start
        # to avoid an io_loop creation before fork
        self.locator = Locator(endpoints=self.locator_endpoints)
        # it's used to reply on `ping` method
        self.locator_status = False

        # active applications
        self.cache = collections.defaultdict(list)
        # routing groups from Locator service
        self.current_rg = {}

        self.logger = logging.getLogger("cocaine.proxy.general")
        self.access_log = logging.getLogger("cocaine.proxy.access")
        self.access_log.propagate = False
        self.logger.info("locators %s",
                         ','.join("%s:%d" % (h, p) for h, p in self.locator_endpoints))

        self.sticky_header = sticky_header

        self.plugins = []
        if srw_config:
            for config in srw_config:
                name, cfg = config["type"], config["args"]
                self.logger.info("initialize plugin %s", name)
                self.plugins.append(load_plugin(name, self, cfg))

        if allow_json_rpc:
            self.plugins.append(load_plugin('cocaine.proxy.jsonrpc.JSONRPC', self, {}))

        self.logger.info("conf path in `%s` configuration service: %s",
                         configuration_service, tracing_conf_path)
        self.unicorn = Service(configuration_service, locator=self.locator)
        self.sampled_apps = {}
        self.default_tracing_chance = default_tracing_chance
        self.tracing_conf_path = tracing_conf_path

        self.io_loop.add_future(self.on_sampling_updates(),
                                lambda x: self.logger.error("the sample updater must not exit"))

        self.timeouts_conf_path = timeouts_conf_path
        self.timeouts = {}
        self.io_loop.add_future(self.on_timeouts_updates(),
                                lambda x: self.logger.error("the timeouts updater must not exit"))

        if request_id_header:
            self.get_request_id = functools.partial(get_request_id, request_id_header,
                                                    force=forcegen_request_header)
        else:
            self.get_request_id = generate_request_id

        # post the watcher for routing groups
        self.io_loop.add_future(self.on_routing_groups_update(),
                                lambda x: self.logger.error("the updater must not exit"))
        # run infinity check locator health status
        self.locator_health_check()

    @gen.coroutine
    def locator_health_check(self, period=5):
        wait_timeot = datetime.timedelta(seconds=period)
        while True:
            try:
                self.logger.debug("check health status of locator via cluster method")
                channel = yield gen.with_timeout(wait_timeot, self.locator.cluster())
                cluster = yield gen.with_timeout(wait_timeot, channel.rx.get())
                self.locator_status = True
                self.logger.debug("dumped cluster %s", cluster)
                yield gen.sleep(period)
            except Exception as err:
                self.logger.error("health status check failed: %s", err)
                self.locator_status = False
                yield gen.sleep(1)

    @gen.coroutine
    def on_routing_groups_update(self):
        uid = gen_uid()
        self.logger.info("generate new unique id %s", uid)
        maximum_timeout = 32  # sec
        timeout = 1  # sec
        while True:
            self.current_rg = {}
            try:
                self.logger.info("subscribe to updates with id %s", uid)
                channel = yield self.locator.routing(uid, True)
                timeout = 1
                while True:
                    new = yield channel.rx.get()
                    if isinstance(new, EmptyResponse):
                        # it means that the cocaine has been stopped
                        self.logger.error("locator sends close")
                        break
                    updates = scan_for_updates(self.current_rg, new)
                    # replace current
                    self.current_rg = new
                    if len(updates) == 0:
                        self.logger.info("locator sends an update message, "
                                         "but no updates have been found")
                        continue

                    self.logger.info("%d routing groups have been refreshed %s",
                                     len(updates), updates)
                    for group in updates:
                        # if we have not created an instance of
                        # the group it is absent in cache
                        if group not in self.cache:
                            self.logger.debug("nothing to update in group %s", group)
                            continue

                        for app in self.cache[group]:
                            self.logger.debug("%s: move %s to the inactive queue to refresh"
                                              " routing group", app.id, app.name)
                            self.migrate_from_cache_to_inactive(app, group)
            except Exception as err:
                timeout = min(timeout << 1, maximum_timeout)
                self.logger.error("error occurred while watching for group updates %s. Sleep %d",
                                  err, timeout)
                yield gen.sleep(timeout)

    @gen.coroutine
    def watch_app(self, name, path):
        version = 0
        self.sampled_apps[name] = self.default_tracing_chance
        try:
            self.logger.info("start watching for sampling updates of %s", name)
            watch_channel = yield self.unicorn.subscribe(path, version)
            while True:
                value, version = yield watch_channel.rx.get()
                self.logger.info("got sampling updates for %s: version %d value %.2f", name, version, value)
                try:
                    weight = float(value)
                    self.sampled_apps[name] = weight
                except ValueError as err:
                    self.logger.error("sample value %s for %s can NOT be converted: %s. Use %f",
                                      value, name, err, self.default_tracing_chance)
                    self.sampled_apps[name] = self.default_tracing_chance
        except ServiceError as err:
            # verify that the err is `zookeeper: no node [-101]``
            if err.code != -101:
                self.logger.error("watching of `%s` raised an unexpected service error (cat. %d): %s", name, err.category, err)
        except Exception as err:
            self.logger.error("watching of %s error: %s", name, err)
        finally:
            self.logger.info("stop watching for sampling updates of %s", name)
            self.sampled_apps.pop(name, None)
            try:
                watch_channel.tx.close()
            except Exception:
                pass

    @gen.coroutine
    def on_sampling_updates(self):
        maximum_timeout = 32  # sec
        timeout = 1  # sec
        listing_version = 0

        while True:
            try:
                listing_channel = yield self.unicorn.children_subscribe(self.tracing_conf_path, listing_version)
                while True:
                    listing_version, apps = yield listing_channel.rx.get()
                    self.logger.info("on_sampling_updates: version %d value %s", listing_version, apps)
                    for app in (i for i in apps if i not in self.sampled_apps):
                        self.watch_app(app, self.tracing_conf_path + "/" + app)
            except Exception as err:
                timeout = min(timeout << 1, maximum_timeout)
                listing_version = 0
                self.logger.error("error occurred while subscribing for sampling updates %s. Sleep %d",
                                  err, timeout)
                yield gen.sleep(timeout)

    @gen.coroutine
    def watch_app_timeouts(self, name, path):
        version = 0
        self.timeouts[name] = {}
        try:
            self.logger.info("start watching for timeouts updates of %s", name)
            watch_channel = yield self.unicorn.subscribe(path, version)
            while True:
                value, version = yield watch_channel.rx.get()
                self.logger.info("got timeouts updates for %s: version %d value %s", name, version, value)
                if isinstance(value, dict):
                    self.timeouts[name] = value
                else:
                    self.logger.error("timeout value %s for %s is not dict", value, name)
                    self.timeouts[name] = {}
        except ServiceError as err:
            # verify that the err is `zookeeper: no node [-101]``
            if err.code != -101:
                self.logger.error("watching of `%s` raised an unexpected service error (cat. %d): %s", name, err.category, err)
        except Exception as err:
            self.logger.error("watching of %s error: %s", name, err)
        finally:
            self.logger.info("stop watching for timeouts updates of %s", name)
            self.timeouts.pop(name, None)
            try:
                watch_channel.tx.close()
            except Exception:
                pass

    @gen.coroutine
    def on_timeouts_updates(self):
        maximum_timeout = 32  # sec
        timeout = 1  # sec
        listing_version = 0

        while True:
            try:
                listing_channel = yield self.unicorn.children_subscribe(self.timeouts_conf_path, listing_version)
                while True:
                    listing_version, apps = yield listing_channel.rx.get()
                    self.logger.info("on_timeouts_updates: version %d value %s", listing_version, apps)
                    for app in (i for i in apps if i not in self.timeouts):
                        self.watch_app_timeouts(app, self.timeouts_conf_path + "/" + app)
            except Exception as err:
                timeout = min(timeout << 1, maximum_timeout)
                listing_version = 0
                self.logger.error("error occurred while subscribing for sampling updates %s. Sleep %d",
                                  err, timeout)
                yield gen.sleep(timeout)

    def get_timeout(self, name, event=''):
        if name in self.timeouts:
            tmts = self.timeouts[name]
            return tmts.get(event) or tmts.get('', DEFAULT_TIMEOUT)

        return DEFAULT_TIMEOUT

    def migrate_from_cache_to_inactive(self, app, name):
        try:
            drop_app_from_cache(self.cache, app, name)
        except Exception as err:
            self.logger.error("app %s %s: drop cache error %s", app, name, err)

        # dispose service after 3 x timeouts
        # assume that all requests will be finished
        self.io_loop.call_later(self.get_timeout(name) * 3,
                                functools.partial(self.dispose, app, name))
        self.logger.info("app %s %s is scheduled to dispose", app, name)

    def move_to_inactive(self, app, name):
        @gen.coroutine
        def wrapper():
            active_apps = len(self.cache[name])
            self.logger.info("%s: preparing to moving %s %s to an inactive queue (active %d)",
                             app.id, app.name, "{0}:{1}".format(*app.address), active_apps)

            try:
                new_app = Service(name, locator=self.locator, timeout=RESOLVE_TIMEOUT)
                self.logger.info("%s: creating an instance of %s", new_app.id, name)
                yield new_app.connect()
                self.logger.info("%s: connect to an app %s endpoint %s ",
                                 new_app.id, new_app.name, "{0}:{1}".format(*new_app.address))
                timeout = (1 + random.random()) * self.refresh_period
                self.io_loop.call_later(timeout, self.move_to_inactive(new_app, name))
                # add to cache only after successfully connected
                self.cache[name].append(new_app)
            except Exception as err:
                self.logger.error("%s: unable to connect to `%s`: %s", new_app.id, name, err)
                # schedule later
                self.io_loop.call_later(self.get_timeout(name), self.move_to_inactive(app, name))
            else:
                self.logger.info("%s: move %s %s to an inactive queue",
                                 app.id, app.name, "{0}:{1}".format(*app.address))
                # current active app will be dropped here
                self.migrate_from_cache_to_inactive(app, name)

        return wrapper

    def dispose(self, app, name):
        self.logger.info("dispose service %s %s", name, app.id)
        app.disconnect()

    def resolve_group_to_version(self, name, value=None):
        """ Pick a version from a routing group using a random or provided value
            A routing group looks like (weight, version):
            {"APP": [[29431330, 'A'], [82426238, 'B'], [101760716, 'C'], [118725487, 'D'], [122951927, 'E']]}
        """
        if name not in self.current_rg:
            return name

        routing_group = self.current_rg[name]
        if len(routing_group) == 0:
            self.logger.warning("empty rounting group %s", name)
            return name

        value = value or random.randint(0, 1 << 32)
        index = upper_bound(routing_group, value)
        return routing_group[index if index < len(routing_group) else 0][1]

    def ping(self, request):
        if self.locator_status:
            fill_response_in(request, httplib.OK, "OK", "OK")
            return

        fill_response_in(request, httplib.SERVICE_UNAVAILABLE,
                         httplib.responses[httplib.SERVICE_UNAVAILABLE],
                         "Failed", proxy_error_headers())

    @context
    @gen.coroutine
    def __call__(self, request):
        for plugin in self.plugins:
            if plugin.match(request):
                request.logger.info('processed by %s plugin', plugin.name())
                try:
                    yield plugin.process(request)
                except PluginNoSuchApplication as err:
                    fill_response_in(request, NO_SUCH_APP, "No such application",
                                     str(err), proxy_error_headers())
                except PluginApplicationError:
                    message = "application error"
                    fill_response_in(request, httplib.INTERNAL_SERVER_ERROR,
                                     httplib.responses[httplib.INTERNAL_SERVER_ERROR],
                                     message, proxy_error_headers())
                except ProxyInvalidRequest:
                    if request.path == "/ping":
                        self.ping(request)
                    else:
                        fill_response_in(request, httplib.NOT_FOUND, httplib.responses[httplib.NOT_FOUND],
                                         "Invalid url", proxy_error_headers())
                except Exception as err:
                    request.logger.exception('plugin %s returned error: %s', plugin.name(), err)
                    message = "unknown error"
                    fill_response_in(request, httplib.INTERNAL_SERVER_ERROR,
                                     httplib.responses[httplib.INTERNAL_SERVER_ERROR],
                                     message, proxy_error_headers())
                return

        try:
            name, event = extract_app_and_event(request)
        except ProxyInvalidRequest:
            if request.path == "/ping":
                self.ping(request)
            else:
                fill_response_in(request, httplib.NOT_FOUND, httplib.responses[httplib.NOT_FOUND],
                                 "Invalid url", proxy_error_headers())
            return

        if getattr(request, "traceid", None) is not None:
            tracing_chance = self.sampled_apps.get(name, self.default_tracing_chance)
            rolled_dice = random.uniform(0, 100)
            request.logger.debug("tracing_chance %f, rolled dice %f", tracing_chance, rolled_dice)
            if tracing_chance < rolled_dice:
                request.logger.info('stop tracing the request')
                request.logger = NULLLOGGER
                request.tracebit = False

        if self.sticky_header in request.headers:
            seed = request.headers.get(self.sticky_header)
            seed_value = header_to_seed(seed)
            request.logger.info('sticky_header has been found: name %s, value %s, seed %d', name, seed, seed_value)
            name = self.resolve_group_to_version(name, seed_value)

        app = yield self.get_service(name, request)

        if app is None:
            message = "current application %s is unavailable" % name
            fill_response_in(request, NO_SUCH_APP, "No Such Application",
                             message, proxy_error_headers(name))
            return

        try:
            # TODO: attempts should be configurable
            yield self.process(request, name, app, event, pack_httprequest(request), self.reelect_app, 2)
        except Exception as err:
            request.logger.exception("error during processing request %s", err)
            fill_response_in(request, httplib.INTERNAL_SERVER_ERROR,
                             httplib.responses[httplib.INTERNAL_SERVER_ERROR],
                             "UID %s: %s" % (request.traceid, str(err)), proxy_error_headers(name))

        request.logger.info("exit from process")

    def info(self):
        return {'services': {'cache': dict(((k, len(v)) for k, v in self.cache.items()))},
                'requests': {'inprogress': self.requests_in_progress,
                             'total': self.requests_total},
                'errors': {'disconnections': self.requests_disconnections},
                'sampling': self.sampled_apps}

    @gen.coroutine
    def reelect_app(self, request, app):
        cache_size = len(self.cache[app.name])
        if cache_size < self.spool_size:
            request.logger.info("spool is not full. Create a new application instance")
            app = yield self.get_service(app.name, request)
        elif cache_size == 1:
            # NOTE: if we have spool_size 1, the same app will be picked
            # Probably we can create a new one and mark the old one inactive
            request.logger.warning("spool size is limited by 1, cannot pick a new instance of th app. Use the old one")
            # pass
        else:
            request.logger.info("pick a random instance of the application")
            try:
                index = self.cache[app.name].index(app)
                request.logger.info("the app is located in cache at pos %d", index)
                if cache_size == 2:  # shortcut
                    picked = (index + 1) % 2
                else:
                    picked = index
                    while picked == index:
                        picked = random.randint(0, cache_size - 1)

                request.logger.info("an instance at pos %d has been picked", index)
                app = self.cache[app.name][picked]
            except ValueError:
                app = random.choice(self.cache[app.name])
        raise gen.Return(app)

    @gen.coroutine
    def process(self, request, name, app, event, data, reelect_app_fn, attempts, timeout=None):
        if timeout is None:
            timeout = self.get_timeout(name, event)
        request.logger.info("start processing event `%s` for an app `%s` (appid: %s) after %.3f ms with timeout %f",
                            event, app.name, app.id, request.request_time() * 1000, timeout)
        parentid = 0

        if request.traceid is not None:
            traceid = int(request.traceid, 16)
            trace = Trace(traceid=traceid, spanid=traceid, parentid=parentid)
        else:
            trace = None

        headers = {
            'trace_bit': '{:d}'.format(request.tracebit),
        }
        if 'authorization' in request.headers:
            headers['authorization'] = request.headers['authorization']

        def on_error(app, err, extra_msg, code=httplib.INTERNAL_SERVER_ERROR):
            if len(extra_msg) > 0 and not extra_msg.endswith(' '):
                extra_msg += ' '
            request.logger.error("%s: %s%s", app.id, extra_msg, err)

            message = "UID %s: application `%s` error: %s" % (request.traceid, app.name, str(err))
            fill_response_in(request, code, httplib.responses[code], message, proxy_error_headers(app.name))

        def check_attempts(app, err):
            if attempts > 0:
                return True
            # we have no attempts more, so quit here
            on_error(app, err, '(no attempts left) ')
            return False

        while attempts > 0:
            attempts -= 1
            processor = None
            try:
                request.logger.debug("%s: enqueue event (attempt %d)", app.id, attempts)
                channel = yield app.enqueue(event, trace=trace, **headers)
                request.logger.debug("%s: send event data (attempt %d)", app.id, attempts)
                yield channel.tx.write(msgpack.packb(data), trace=trace)
                yield channel.tx.close(trace=trace)
                request.logger.debug("%s: waiting for a code and headers (attempt %d)",
                                     app.id, attempts)
                code_and_headers = yield channel.rx.get(timeout=timeout)
                request.logger.debug("%s: code and headers have been received (attempt %d)",
                                     app.id, attempts)
                code, raw_headers = msgpack.unpackb(code_and_headers)
                headers = httputil.HTTPHeaders(raw_headers)

                cocaine_http_proto_version = headers.get(X_COCAINE_HTTP_PROTO_VERSION)
                if cocaine_http_proto_version is None or cocaine_http_proto_version == "1.0":
                    cocaine_http_proto_version = "1.0"

                    def stop_condition(body):
                        return isinstance(body, EmptyResponse)
                elif cocaine_http_proto_version == "1.1":
                    def stop_condition(body):
                        return isinstance(body, EmptyResponse) or len(body) == 0
                else:
                    raise Exception("unsupported X-Cocaine-HTTP-Proto-Version: %s" % cocaine_http_proto_version)

                processor = BodyProcessor.make_processor(
                    headers.get('Content-Length'),
                    request, name, code, headers)

                while True:
                    body = yield channel.rx.get(timeout=timeout)
                    if stop_condition(body):
                        request.logger.info("%s: body finished (attempt %d)", app.id, attempts)
                        break

                    request.logger.debug("%s: received %d bytes as a body chunk (attempt %d)",
                                         app.id, len(body), attempts)

                    processor.swallow(body)

            except gen.TimeoutError as err:
                on_error(app, err, '', httplib.GATEWAY_TIMEOUT)

            except (DisconnectionError, StreamClosedError) as err:
                self.requests_disconnections += 1
                # Probably it's dangerous to retry requests all the time.
                # I must find the way to determine whether it failed during writing
                # or reading a reply. And retry only writing fails.
                request.logger.error("%s: %s", app.id, err)
                if not check_attempts(app, err):
                    return

                # Seems on_close callback is not called in case of connecting through IPVS
                # We detect disconnection here to avoid unnecessary errors.
                # Try to reconnect here and give the request a go
                try:
                    start_time = time.time()
                    reconn_timeout = timeout - request.request_time()
                    request.logger.info("%s: connecting with timeout %.fms", app.id, reconn_timeout * 1000)
                    yield gen.with_timeout(start_time + reconn_timeout, app.connect(request.traceid))
                    reconn_time = time.time() - start_time
                    request.logger.info("%s: connecting took %.3fms", app.id, reconn_time * 1000)
                except Exception as err:
                    request.logger.error("%s: unable to reconnect: %s (%d attempts left)", err, attempts)
                # We have an attempt to process request again.
                # Jump to the begining of `while attempts > 0`, either we connected successfully
                # or we were failed to connect
                continue

            except ServiceError as err:
                if not check_attempts(app, err):
                    return

                # if the application has been restarted, we get broken pipe code
                # and system category
                if err.category in SYSTEMCATEGORY and err.code == EAPPSTOPPED:
                    request.logger.error("%s: the application has been restarted", app.id)
                    app.disconnect()
                    continue

                elif err.category in OVERSEERCATEGORY and err.code == EQUEUEISFULL:
                    request.logger.error("%s: queue is full. Pick another application instance", app.id)
                    try:
                        app = yield reelect_app_fn(request, app)
                    except Exception as reelect_err:
                        on_error(app, reelect_err, '(could not reelect app)')
                        return
                    request.logger.info("fetched new app from reelect_app_fn")
                    continue

                on_error(app, err, '')

            except Exception as err:
                on_error(app, err, '(unknown error) ')

            else:
                if processor:
                    processor.finish()

            # to return from all errors except Disconnection
            # or receiving a good reply
            return

    @gen.coroutine
    def get_service(self, name, request):
        # cache isn't full for the current application
        if len(self.cache[name]) < self.spool_size:
            logger = request.logger
            try:
                app = Service(name, locator=self.locator, timeout=RESOLVE_TIMEOUT)
                logger.info("%s: creating an instance of %s", app.id, name)
                self.cache[name].append(app)
                yield app.connect(request.traceid)
                logger.info("%s: connect to an app %s endpoint %s ",
                            app.id, app.name, "{0}:{1}".format(*app.address))

                timeout = (1 + random.random()) * self.refresh_period
                self.io_loop.call_later(timeout, self.move_to_inactive(app, name))
            except Exception as err:
                logger.error("%s: unable to connect to `%s`: %s", app.id, name, err)
                drop_app_from_cache(self.cache, app, name)
                raise gen.Return()
            else:
                raise gen.Return(app)

        # get an instance from cache
        chosen = random.choice(self.cache[name])
        raise gen.Return(chosen)


def enable_logging(options):
    if options.logging is None or options.logging.lower() == "none":
        return

    general_logger = logging.getLogger("cocaine.proxy.general")
    general_logger.setLevel(getattr(logging, options.logging.upper()))
    general_formatter = logging.Formatter(options.generallogfmt, datefmt=options.datefmt)

    access_logger = logging.getLogger("cocaine.proxy.access")
    access_logger.setLevel(getattr(logging, options.logging.upper()))
    access_formatter = logging.Formatter(options.accesslogfmt, datefmt=options.datefmt)

    cocainelogger = None
    if options.logframework:
        cocainelogger = logging.getLogger("cocaine.baseservice")
        cocainelogger.setLevel(getattr(logging, options.logging.upper()))

    if options.log_to_cocaine:
        Logger().target = "tornado-proxy"
        handler = CocaineHandler()
        general_logger.addHandler(handler)
        if cocainelogger:
            cocainelogger.addHandler(handler)

        access_logger.addHandler(handler)

    if options.log_file_prefix:
        handler = logging.handlers.WatchedFileHandler(
            filename=options.log_file_prefix,
        )
        handler.setFormatter(general_formatter)
        general_logger.addHandler(handler)

        handler = logging.handlers.WatchedFileHandler(
            filename=options.log_file_prefix,
        )
        handler.setFormatter(access_formatter)
        access_logger.addHandler(handler)

        if cocainelogger:
            cocainehandler = logging.handlers.WatchedFileHandler(
                filename=options.log_file_prefix + "framework.log"
            )
            cocainehandler.setFormatter(general_formatter)
            cocainelogger.addHandler(cocainehandler)

    if options.log_to_stderr or (options.log_to_stderr is None and not general_logger.handlers):
        stderr_handler = logging.StreamHandler()
        stderr_handler.setFormatter(general_formatter)

        general_logger.addHandler(stderr_handler)
        if cocainelogger:
            cocainelogger.addHandler(stderr_handler)

        stderr_handler = logging.StreamHandler()
        stderr_handler.setFormatter(access_formatter)
        access_logger.addHandler(stderr_handler)


def enable_gc_stats():
    import gc
    gc.set_debug(gc.DEBUG_STATS)


def show_version(dummy):
    print("cocaine tools & proxy: %s" % tools_version)
    sys.exit(0)


def main():
    from tornado import options

    default_general_logformat = "[%(asctime)s.%(msecs)d]\t[%(filename).5s:%(lineno)d]\t%(levelname)s\t%(message)s"
    default_access_logformat = "[%(asctime)s.%(msecs)d]\t[%(filename).5s:%(lineno)d]\t%(levelname)s\t%(trace_id)s\t%(message)s"

    opts = options.OptionParser()
    opts.define("version", type=bool, help="show version and exit", callback=show_version)
    opts.define("locators", default=["localhost:10053"],
                type=str, multiple=True, help="comma-separated endpoints of locators")
    opts.define("cache", default=DEFAULT_SERVICE_CACHE_COUNT,
                type=int, help="count of instances per service")
    opts.define("config", help="path to configuration file", type=str,
                callback=lambda path: opts.parse_config_file(path, final=False))
    opts.define("count", default=1, type=int, help="count of tornado processes")
    opts.define("endpoints", default=["tcp://localhost:8080"], type=str, multiple=True,
                help="Specify endpoints to bind on: prefix unix:// or tcp:// should be used")
    opts.define("request_header", default="X-Request-Id", type=str,
                help="header used as a trace id")
    opts.define("forcegen_request_header", default=False, type=bool,
                help="enable force generation of the request header")
    opts.define("sticky_header", default="X-Cocaine-Sticky", type=str, help="sticky header name")
    opts.define("gcstats", default=False, type=bool, help="print garbage collector stats to stderr")
    opts.define("srwconfig", default="", type=str, help="path to srwconfig")
    opts.define("allow_json_rpc", default=True, type=bool, help="allow JSON RPC module")

    # tracing options
    opts.define("tracing_chance", default=DEFAULT_TRACING_CHANCE,
                type=float, help="default chance for an app to be traced")
    opts.define("configuration_service", default="unicorn",
                type=str, help="name of configuration service")
    opts.define("tracing_conf_path", default="/zipkin_sampling",
                type=str, help="path to the configuration nodes in the configuration service")

    # various logging options
    opts.define("logging", default="info",
                help=("Set the Python log level. If 'none', tornado won't touch the "
                      "logging configuration."), metavar="debug|info|warning|error|none")
    opts.define("log_to_cocaine", default=False, type=bool, help="log to cocaine")
    opts.define("log_to_stderr", type=bool, default=None,
                help=("Send log output to stderr. "
                      "By default use stderr if --log_file_prefix is not set and "
                      "no other logging is configured."))
    opts.define("log_file_prefix", type=str, default=None, metavar="PATH",
                help=("Path prefix for log file"))
    opts.define("datefmt", type=str, default="%z %d/%b/%Y:%H:%M:%S", help="datefmt")
    opts.define("generallogfmt", type=str, help="log format of general logging system",
                default=default_general_logformat)
    opts.define("accesslogfmt", type=str, help="log format of access logging system",
                default=default_access_logformat)
    opts.define("logframework", type=bool, default=False,
                help="enable logging various framework messages")

    # util server
    opts.define("utilport", default=8081, type=int, help="listening port number for an util server")
    opts.define("utiladdress", default="127.0.0.1", type=str, help="address for an util server")
    opts.define("enableutil", default=False, type=bool, help="enable util server")
    opts.parse_command_line()

    srw_config = None
    if opts.srwconfig:
        try:
            srw_config = load_srw_config(opts.srwconfig)
        except Exception as err:
            print("unable to load SRW config: %s" % err)
            exit(1)

    use_reuseport = hasattr(socket, "SO_REUSEPORT")
    endpoints = Endpoints(opts.endpoints)
    sockets = []

    for path in endpoints.unix:
        sockets.append(bind_unix_socket(path, mode=0o666))

    if not use_reuseport:
        for endpoint in endpoints.tcp:
            # We have to bind before fork to distribute sockets to our forks
            socks = bind_sockets(endpoint.port, address=endpoint.host)
            sockets.extend(socks)

    if opts.enableutil:
        utilsockets = bind_sockets(opts.utilport, address=opts.utiladdress)

    try:
        if opts.count != 1:
            process.fork_processes(opts.count)

        enable_logging(opts)

        if opts.gcstats:
            enable_gc_stats()

        if use_reuseport:
            for endpoint in endpoints.tcp:
                # We have to bind before fork to distribute sockets to our forks
                socks = bind_sockets(endpoint.port, address=endpoint.host, reuse_port=True)
                sockets.extend(socks)

        proxy = CocaineProxy(locators=opts.locators, cache=opts.cache,
                             request_id_header=opts.request_header,
                             sticky_header=opts.sticky_header,
                             forcegen_request_header=opts.forcegen_request_header,
                             default_tracing_chance=opts.tracing_chance,
                             srw_config=srw_config,
                             allow_json_rpc=opts.allow_json_rpc)
        server = HTTPServer(proxy)
        server.add_sockets(sockets)

        if opts.enableutil:
            utilsrv = HTTPServer(UtilServer(proxy=proxy))
            utilsrv.add_sockets(utilsockets)

        tornado.ioloop.IOLoop.current().start()
    except KeyboardInterrupt:
        pass


if __name__ == '__main__':
    main()



from tornado import gen

from cocaine.decorators import coroutine

from . import SecurePlugin


class Promiscuous(SecurePlugin):
    def __init__(self, repo):
        super(Promiscuous, self).__init__(repo)

    def ty(self):
        return ''

    @coroutine
    def fetch_token(self):
        raise gen.Return('')



#
#    Copyright (c) 2013+ Anton Tyurin <noxiouz@yandex.ru>
#    Copyright (c) 2011-2013 Other contributors as noted in the AUTHORS file.
#
#    This file is part of Cocaine.
#
#    Cocaine is free software; you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    Cocaine is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import os

DEFAULT_EDITOR_PATH = "/usr/bin/editor"


def locate_editor():
    editor = None
    # check /usr/bin/editor
    if os.path.exists(DEFAULT_EDITOR_PATH):
        editor = DEFAULT_EDITOR_PATH

    # check EDITOR variable
    editor = os.environ.get('EDITOR') or editor

    # check VISUAL variable
    editor = os.environ.get('VISUAL') or editor
    return editor



#
# Copyright (c) 2013+ Anton Tyurin <noxiouz@yandex.ru>
# Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
# Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
# This file is part of Cocaine-tools.
#
# Cocaine is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Cocaine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import datetime
import itertools
import time
from operator import itemgetter

from tornado import gen

from cocaine.tools import actions, log
from cocaine.decorators import coroutine
from cocaine.tools.actions import app

__author__ = 'Evgeny Safronov <division494@gmail.com>'

index_format = 'cocaine-%Y-%m-%d'


def parse_crashlog_day_format(day_string):
    if not day_string:
        return day_string

    if 'today'.startswith(day_string):
        return datetime.date.today().strftime(index_format)
    elif 'yesterday'.startswith(day_string):
        yesterday = datetime.date.today() - datetime.timedelta(days=1)
        return yesterday.strftime(index_format)
    else:
        values_count = day_string.count("-")
        if values_count == 0:  # only day specified
            today = datetime.date.today()
            day = datetime.datetime.strptime(day_string, "%d").replace(year=today.year,
                                                                       month=today.month)
            return day.strftime(index_format)
        elif values_count == 1:  # day and month
            day = datetime.datetime.strptime(day_string,
                                             "%d-%m").replace(year=datetime.date.today().year)
            return day.strftime(index_format)
        elif values_count == 2:  # the whole date
            return datetime.datetime.strptime(day_string, "%d-%m-%Y").strftime(index_format)
    raise ValueError("Invalid day format %s. Must be day-month-year|today|yesterday" % day_string)


def days_range(from_date, to_date, delta=datetime.timedelta(days=1)):
    fdate = datetime.datetime.strptime(from_date, "%Y-%m-%d")
    tdate = datetime.datetime.strptime(to_date, "%Y-%m-%d")
    while fdate <= tdate:
        yield fdate
        fdate = fdate + delta
    return


class List(actions.Storage):
    def __init__(self, storage, name, day_string=''):
        super(List, self).__init__(storage)
        self.name = name
        if not self.name:
            raise ValueError('Please specify a crashlog name')
        self.day = parse_crashlog_day_format(day_string)

    @coroutine
    def execute(self):
        indexes = [self.name]
        if self.day:
            indexes.append(self.day)
        channel = yield self.storage.find('crashlogs', indexes)
        listing = yield channel.rx.get()
        raise gen.Return(listing)


def _parseCrashlogs(crashlogs, timestamp=None):
    def is_filter(arg):
        return arg == timestamp if timestamp else True

    _list = (log.split(':', 1) for log in crashlogs)
    return [(ts, time.ctime(float(ts) / 1000000), name) for ts, name in _list if is_filter(ts)]


class Specific(actions.Storage):
    def __init__(self, storage, name, timestamp=None):
        super(Specific, self).__init__(storage)
        self.name = name
        self.timestamp = timestamp
        if not self.name:
            raise ValueError('Please specify application name')


class View(Specific):
    @coroutine
    def execute(self):
        channel = yield self.storage.find('crashlogs', [self.name])
        crashlogs = yield channel.rx.get()
        parsed_crashlogs = _parseCrashlogs(crashlogs, timestamp=self.timestamp)
        contents = []
        if not self.timestamp and parsed_crashlogs:
            parsed_crashlogs = [max(parsed_crashlogs, key=itemgetter(1))]
        for crashlog in parsed_crashlogs:
            key = '%s:%s' % (crashlog[0], crashlog[2])
            channel = yield self.storage.read('crashlogs', key)
            content = yield channel.rx.get()
            contents.append(content)
        raise gen.Return(''.join(contents))


class Remove(Specific):
    @coroutine
    def execute(self):
        channel = yield self.storage.find('crashlogs', [self.name])
        crashlogs = yield channel.rx.get()
        parsed_crashlogs = _parseCrashlogs(crashlogs, timestamp=self.timestamp)
        for crashlog in parsed_crashlogs:
            try:
                key = '%s:%s' % (crashlog[0], crashlog[2])
                channel = yield self.storage.remove('crashlogs', key)
                yield channel.rx.get()
            except Exception as err:
                log.error("unable to delete crashlog %s: %s", str(crashlog), err)
        raise gen.Return('Done')


class RemoveAll(Remove):
    def __init__(self, storage, name):
        super(RemoveAll, self).__init__(storage, name, timestamp=None)


class Status(actions.Storage):
    @coroutine
    def execute(self):
        applications = yield app.List(self.storage).execute()
        crashed = []
        for application in applications:
            crashlogs = yield List(self.storage, application).execute()
            if crashlogs:
                last = max(_parseCrashlogs(crashlogs), key=lambda (timestamp, time, uuid): timestamp)
                crashed.append((application, last, len(crashlogs)))
        raise gen.Return(crashed)


def splitted(collection, sep=None, maxsplit=None):
    for item in collection:
        yield item.split(sep, maxsplit)


def filtered(crashlogs):
    for (ts, uuid) in splitted(crashlogs, ':', 1):
        yield int(ts), uuid


class Clean(Specific):
    def __init__(self, storage, name, size, timestamp=None):
        super(Clean, self).__init__(storage, name, timestamp)
        self.size = int(size)

    @coroutine
    def execute(self):
        if not self.name:
            apps = yield app.List(self.storage).execute()
        else:
            apps = [self.name]

        result = []
        if self.timestamp:
            try:
                dt = datetime.datetime.strptime(self.timestamp, '%Y-%m-%dT%H:%M:%S')
                timestamp = int(time.mktime(dt.timetuple())) * 1000000 + dt.microsecond
            except ValueError:
                timestamp = int(self.timestamp)

            for app_name in apps:
                channel = yield self.storage.find('crashlogs', [app_name])
                crashlogs = yield channel.rx.get()
                result = filter(lambda (ts, uuid): ts < timestamp, filtered(crashlogs))
        elif self.size > 0:
            for app_name in apps:
                channel = yield self.storage.find('crashlogs', [app_name])
                crashlogs = yield channel.rx.get()
                result = itertools.islice(
                    sorted(filtered(crashlogs[0]), key=lambda (ts, uuid): ts, reverse=True), self.size, None)

        for crashlog in result:
            print('removing', '%d:%s' % crashlog)
            channel = yield self.storage.remove('crashlogs', '%d:%s' % crashlog)
            yield channel.rx.get()
        raise gen.Return('Done')


class CleanRange(object):
    def __init__(self, storage, from_day, to_day="yesterday"):
        self.storage = storage
        if not from_day:
            raise ValueError("from-day must have a value")
        # strip cocaine-
        self.from_day = parse_crashlog_day_format(from_day)[len("cocaine-"):]
        self.to_day = parse_crashlog_day_format(to_day)[len("cocaine-"):]

    @coroutine
    def execute(self):
        listing = list()
        for day in days_range(self.from_day, self.to_day):
            tag = day.strftime(index_format)
            items = yield (yield self.storage.find('crashlogs', [tag])).rx.get()
            log.info("found %d crashlog(s) for %s", len(items), tag)
            listing.extend(items)

        log.info("there are %d crashlog(s)", len(listing))
        step = len(listing) / 100
        for i, key in enumerate(listing, start=1):
            try:
                if not i % step:
                    log.info("(%d/%d) %d%% of crashlogs have been removed", i, len(listing), i / step)
                yield (yield self.storage.remove('crashlogs', key)).rx.get()
            except Exception as err:
                log.error("unable to remove %s, %s", key, err)



import json
import os

import click
from cerberus import Validator
from tornado import gen

from cocaine.decorators import coroutine
from cocaine.exceptions import ServiceError

from cocaine.tools.actions import Action


_PREFIX = '/.acls/event'

ERROR_CATEGORY_UNICORN = 16639
ERROR_CODE_NO_NODE = -101

VERSION_NOT_EXISTS = -1


class List(Action):
    def __init__(self, unicorn):
        self._unicorn = unicorn

    @coroutine
    def execute(self):
        path = os.path.join(_PREFIX)
        channel = yield self._unicorn.children_subscribe(path)
        version, services = yield channel.rx.get()
        raise gen.Return(services)


class View(Action):
    def __init__(self, service, unicorn):
        self._service = service
        self._unicorn = unicorn

    @coroutine
    def execute(self):
        raise gen.Return({
            'cids': (yield self._walk('cids')),
            'uids': (yield self._walk('uids')),
        })

    @coroutine
    def _walk(self, ty):
        path = os.path.join(_PREFIX, self._service, ty)

        try:
            channel = yield self._unicorn.children_subscribe(path)
            version, ids = yield channel.rx.get()
        except ServiceError as err:
            if err.category == ERROR_CATEGORY_UNICORN and err.code == ERROR_CODE_NO_NODE:
                raise gen.Return({})
            else:
                raise err

        content = {}
        for id_ in ids:
            channel = yield self._unicorn.get(os.path.join(_PREFIX, self._service, ty, id_))
            value, version = yield channel.rx.get()

            content[id_] = value
        raise gen.Return(content)


class Add(Action):
    def __init__(self, service, event, scope, id_, unicorn):
        self._service = service
        self._event = event
        self._path = os.path.join(_PREFIX, self._service, scope, '{}'.format(id_))
        self._unicorn = unicorn

    @coroutine
    def execute(self):
        channel = yield self._unicorn.get(self._path)
        value, version = yield channel.rx.get()

        if version == VERSION_NOT_EXISTS:
            channel = yield self._unicorn.create(self._path, {
                self._event: {}
            })
        else:
            # Do not override if already exists.
            if self._event not in value:
                value[self._event] = {}
            channel = yield self._unicorn.put(self._path, value, version)

        yield channel.rx.get()


class AddUser(Add):
    def __init__(self, service, event, uid, unicorn):
        super(AddUser, self).__init__(service, event, 'uids', uid, unicorn)


class AddClient(Add):
    def __init__(self, service, event, cid, unicorn):
        super(AddClient, self).__init__(service, event, 'cids', cid, unicorn)


class AddBoth(Action):
    def __init__(self, service, event, cids, uids, unicorn):
        self._service = service
        self._cids = cids
        self._uids = uids
        self._event = event
        self._unicorn = unicorn

    @coroutine
    def execute(self):
        for uid in self._uids:
            yield AddUser(self._service, self._event, uid, self._unicorn).execute()
        for cid in self._cids:
            yield AddClient(self._service, self._event, cid, self._unicorn).execute()


class Remove(Action):
    def __init__(self, service, event, scope, id_, unicorn):
        self._service = service
        self._event = event
        self._path = os.path.join(_PREFIX, self._service, scope, '{}'.format(id_))
        self._unicorn = unicorn

    @coroutine
    def execute(self):
        channel = yield self._unicorn.get(self._path)
        value, version = yield channel.rx.get()

        print(value, version)
        if version != VERSION_NOT_EXISTS:
            value.pop(self._event)

            if len(value) == 0:
                channel = yield self._unicorn.remove(self._path, VERSION_NOT_EXISTS)
            else:
                channel = yield self._unicorn.put(self._path, value, version)

        yield channel.rx.get()


class RemoveUser(Remove):
    def __init__(self, service, event, uid, unicorn):
        super(RemoveUser, self).__init__(service, event, 'uids', uid, unicorn)


class RemoveClient(Remove):
    def __init__(self, service, event, cid, unicorn):
        super(RemoveClient, self).__init__(service, event, 'cids', cid, unicorn)


class Edit(Action):
    SUB_SCHEMA = {
        'type': 'dict',
        'required': True,
        'propertyschema': {
            'type': 'string',
        },
        'valueschema': {
            'type': 'dict',
            'propertyschema': {
                'type': 'string',
            },
            'valueschema': {
                'type': 'dict',
            },
        }
    }

    SCHEMA = {
        'cids': SUB_SCHEMA,
        'uids': SUB_SCHEMA,
    }

    EMPTY_CONTENT = {
        'cids': {},
        'uids': {},
    }

    def __init__(self, service, unicorn):
        self._service = service
        self._unicorn = unicorn
        self._validator = Validator(self.SCHEMA)

    @coroutine
    def execute(self):
        try:
            content = yield View(self._service, self._unicorn).execute()
        except ServiceError as err:
            # Create default empty ACL if not exists.
            if err.category == ERROR_CATEGORY_UNICORN and err.code == ERROR_CODE_NO_NODE:
                content = self.EMPTY_CONTENT
            else:
                raise err

        updated = click.edit(json.dumps(content, indent=4), require_save=False)
        updated = json.loads(updated)
        self._validator.validate(updated)
        if self._validator.errors:
            raise ValueError('failed to validate to ACLs: {}'.format(self._validator.errors))

        for scope in ('uids', 'cids'):
            # Create new records.
            for id_ in set(updated[scope]) - set(content[scope]):
                path = os.path.join(_PREFIX, self._service, scope, '{}'.format(id_))
                channel = yield self._unicorn.create(path, updated[scope][id_])
                yield channel.rx.get()
            # Remove old records.
            for id_ in set(content[scope]) - set(updated[scope]):
                path = os.path.join(_PREFIX, self._service, scope, '{}'.format(id_))
                channel = yield self._unicorn.remove(path, VERSION_NOT_EXISTS)
                yield channel.rx.get()

            # Update non-changed records with changed values.
            for id_ in set(updated[scope]) & set(content[scope]):
                if updated[scope][id_] != content[scope][id_]:
                    path = os.path.join(_PREFIX, self._service, scope, '{}'.format(id_))
                    channel = yield self._unicorn.get(path)
                    value, version = yield channel.rx.get()
                    channel = yield self._unicorn.put(path, updated[scope][id_], version)
                    yield channel.rx.get()



# coding=utf-8

import contextlib
import copy
import logging
import os

import click
from click import Choice
import collections
import six
import yaml
from tornado import gen
from tornado.util import import_object
from cerberus import Validator

from cocaine.tools.cli import Executor

from cocaine.decorators import coroutine
from cocaine.exceptions import CocaineError
from cocaine.services import Locator, Service
from .plugins.secure.promiscuous import Promiscuous
from .version import __version__


CONFIG_GLOB = '/etc/cocaine/.cocaine/tools.yml'
CONFIG_USER = '~/.cocaine/tools.yml'
CONFIG_PATHS = [CONFIG_GLOB, CONFIG_USER]

DEFAULT_LOCATOR_HOST = 'localhost'
DEFAULT_LOCATOR_PORT = 10053

log = logging.getLogger('cocaine.tools')


def _print_experimental_warning():
    click.echo('')
    click.echo('THIS COMMAND IS EXPERIMENTAL. DO NOT DEPEND ON IT IN YOUR SCRIPTS')
    click.echo('')


class SecureServiceError(CocaineError):
    pass


def set_verbosity(ctx, param, value):
    levels = [
        logging.NOTSET,
        logging.ERROR,
        logging.WARN,
        logging.INFO,
        logging.DEBUG
    ]

    loggers = ['cocaine.tools']

    # Enable cocaine-framework logging if there are more than maximum verbosity requested.
    if value > 4:
        value = 4
        loggers.append('cocaine')

    level = levels[value]

    formatter = logging.Formatter('%(levelname)-1.1s, %(asctime)s: %(message)s')

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    handler.setLevel(level)
    for name in loggers:
        lg = logging.getLogger(name)
        lg.propagate = False
        lg.addHandler(handler)
        lg.setLevel(level)
    return value


_global_options = [
    click.option('-v', count=True, callback=set_verbosity, help='Enable additional output.'),
    click.option('--host', metavar='', help='Locator hostname.'),
    click.option('--port', metavar='', help='Locator port.'),
    click.option('--timeout', metavar='', default=20, help='Operation timeout.'),
]


@contextlib.contextmanager
def watch_progress(fmt, *args):
    message = (fmt % args).capitalize()
    try:
        click.echo('[ ] {}\033[1A'.format(message))
        yield
    except Exception:
        click.echo('[{}] {}'.format(click.style('✘', fg='red'), message))
        raise
    else:
        click.echo('[{}] {}'.format(click.style('✔', fg='green'), message))


def with_options(func):
    for option in reversed(_global_options):
        func = option(func)
    return func


class SecureService(object):
    def __init__(self, secure, wrapped):
        self._secure = secure
        self._wrapped = wrapped

    @coroutine
    def connect(self, traceid=None):
        yield self._wrapped.connect(traceid)

    def disconnect(self):
        return self._wrapped.disconnect()

    def __getattr__(self, name):
        @coroutine
        def wrapper(*args, **kwargs):
            try:
                kwargs['authorization'] = yield self._secure.fetch_token()
            except Exception as err:
                raise SecureServiceError('failed to fetch secure token: {}'.format(err))
            raise gen.Return((yield getattr(self._wrapped, name)(*args, **kwargs)))
        return wrapper


class ServiceFactory(object):
    def create_service(self, name):
        raise NotImplementedError

    def create_secure_service(self, name):
        raise NotImplementedError


class PooledServiceFactory(ServiceFactory):
    def __init__(self, endpoints):
        self._secure = None
        self._endpoints = endpoints
        self._cache = {}

    @property
    def secure(self):
        return self._secure

    @secure.setter
    def secure(self, value):
        self._secure = value

    def create_service(self, name):
        if name not in self._cache:
            if name == 'locator':
                service = Locator(endpoints=self._endpoints)
            else:
                service = Service(name, endpoints=self._endpoints)
            self._cache[name] = service
        return self._cache[name]

    def create_secure_service(self, name):
        return SecureService(self._secure, self.create_service(name))


class Configurator(object):
    SCHEMA = {
        'locator': {
            'type': 'dict',
            'schema': {
                'host': {
                    'type': 'string',
                },
                'port': {
                    'type': 'integer',
                    'min': 0,
                    'max': 65535,
                }
            }
        },
        'secure': {
            'type': 'dict',
            'schema': {
                'mod': {
                    'type': 'string',
                    'allowed': ['TVM'],
                },
                'client_id': {
                    'type': 'integer',
                },
                'client_secret': {
                    'type': 'string',
                }
            }
        }
    }

    def __init__(self):
        self._config = {}
        self._validator = Validator(self.SCHEMA)

    @property
    def config(self):
        return self._config

    def update(self, paths=None):
        if paths is None:
            paths = CONFIG_PATHS

        log.debug('scanning configs at %s', paths)
        configs = []
        for path in paths:
            if os.path.exists(os.path.expanduser(path)):
                content = yaml.safe_load(open(os.path.expanduser(path)).read())
                if content is None:
                    log.debug('config %s is empty, skipping', path)
                    continue
                configs.append((path, content))
                log.debug('successfully read config from %s\n%s', path, content)
            else:
                log.debug('config %s was not found, skipping', path)

        self._config.clear()
        if len(configs) == 0:
            log.info('no config found - use default values')
        else:
            used = []
            for filename, config in configs:
                used.append(filename)
                self._config = Configurator._merge_dicts(self._config, config)
            log.info('loaded config(s) from %s', used)
        self._validate()

    @staticmethod
    def _merge_dicts(src, d):
        key = None
        try:
            if src is None or isinstance(src, (six.string_types, float, six.integer_types)):
                # Border case for first run or if `src` is a primitive.
                src = d
            elif isinstance(src, list):
                # Lists can be only appended.
                if isinstance(d, list):
                    src.extend(d)
                else:
                    src.append(d)
            elif isinstance(src, dict):
                # Dicts must be merged.
                if isinstance(d, dict):
                    for key in d:
                        if key in src:
                            src[key] = Configurator._merge_dicts(src[key], d[key])
                        else:
                            src[key] = d[key]
                else:
                    raise RuntimeError('Cannot merge non-dict "%s" into dict "%s"' % (d, src))
            else:
                raise RuntimeError('NOT IMPLEMENTED "%s" into "%s"' % (d, src))
        except TypeError as e:
            raise RuntimeError(
                'TypeError "%s" in key "%s" when merging "%s" into "%s"' % (e, key, d, src))
        return src

    def _validate(self):
        self._validator.validate(self._config)

        if self._validator.errors:
            raise ValueError('failed to validate configuration file: {}'.format(self._validator.errors))


class PluginLoader(object):
    def __init__(self):
        self._secure = Promiscuous(None)

    def load(self, config, repo):
        self._load_secure(config, repo)

    def _load_secure(self, config, repo):
        if 'secure' in config:
            ty = config['secure'].get('mod')
            if ty is None:
                return
            kwargs = copy.deepcopy(config['secure'])
            kwargs.pop('mod')

            if not ty.startswith('cocaine.tools.plugins.secure'):
                ty = 'cocaine.tools.plugins.secure.' + ty.lower() + '.' + ty
            self._secure = import_object(ty)(repo, **kwargs)
            log.info('imported "%s" secure plugin', ty)

    def secure(self):
        return self._secure


class Context(object):
    def __init__(self, host, port, timeout, **kwargs):
        self._timeout = timeout
        self._options = kwargs

        self._configurator = Configurator()
        self._configurator.update()

        if host is None:
            host = self._configurator.config.get('locator', {}).get('host', DEFAULT_LOCATOR_HOST)

        if port is None:
            port = self._configurator.config.get('locator', {}).get('port', DEFAULT_LOCATOR_PORT)

        self._endpoints = [(host, int(port))]

        self._repo = PooledServiceFactory(endpoints=self._endpoints)

        self._loader = PluginLoader()
        self._loader.load(self._configurator.config, self._repo)

        self._repo.secure = self._loader.secure()

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, value):
        self._timeout = value

    @property
    def repo(self):
        return self._repo

    @property
    def locator(self):
        return self._repo.create_service('locator')

    def execute_action(self, __name, **kwargs):
        return Executor(timeout=self.timeout).execute_action(__name, **kwargs)


ALIASES = {
    'i': 'info',
}


class AliasedGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        cmd = ALIASES.get(cmd_name)
        if cmd is None:
            return click.Group.get_command(self, ctx, cmd_name)
        else:
            return click.Group.get_command(self, ctx, cmd)


@click.group(cls=AliasedGroup)
def tools():
    pass


@tools.command()
def version():
    """
    Show version.
    """
    click.echo(__version__)


@tools.command()
@click.option('-n', '--name', metavar='', required=True, help='Service name.')
@with_options
def locate(name, **kwargs):
    """
    Show resolve information about specified service.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('locate', **{
        'name': name,
        'locator': ctx.locator,
    })


@tools.command()
@click.option('-n', '--name', metavar='', required=True, help='Routing group name.')
@with_options
def routing(name, **kwargs):
    """
    Show information about the requested routing group.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('routing', **{
        'name': name,
        'locator': ctx.locator,
    })


@tools.command()
@click.option('--resolve', metavar='', default=False, help='Show IPs instead of hostname.')
@with_options
def cluster(resolve, **kwargs):
    """
    Show cluster info.
    """
    # Actually we have IPs and we need not do anything to resolve them to IPs. So the default
    # behavior fits better to this option name.

    ctx = Context(**kwargs)
    ctx.execute_action('cluster', **{
        'locator': ctx.locator,
        'resolve': resolve,
    })


@tools.command()
@click.option('-n', '--name', metavar='', help='Application name.')
@click.option('-m', is_flag=True, default=False, help='Expand manifest.')
@click.option('-p', is_flag=True, default=False, help='Expand profile.')
@click.option('-b', is_flag=True, default=False, help='Show only brief info (disables -p and -m).')
@click.option('-w', is_flag=True, default=True, help='Do not use wildcard to match app name.')
@with_options
def info(name, m, p, b, w, **kwargs):
    """
    Show information about cocaine runtime.

    Return json-like string with information about cocaine-runtime.

    If the name option is not specified, shows information about all applications. Flags can be
    specified for fine-grained control of the output verbosity.
    """
    m = (m << 1) & 0b010
    p = (p << 2) & 0b100

    # Brief disables all further flags.
    if b:
        flags = 0b000
    else:
        flags = m | p | 0b001

    ctx = Context(**kwargs)
    ctx.execute_action('info', **{
        'node': ctx.repo.create_secure_service('node'),
        'locator': ctx.locator,
        'name': name,
        'flags': flags,
        'use_wildcard': w,
        'timeout': ctx.timeout,
    })


@tools.group(name='app')
def app_group():
    """
    Application commands.
    """
    pass


@tools.group(name='profile')
def profile_group():
    """
    Profile commands.
    """
    pass


@tools.group(name='runlist')
def runlist_group():
    """
    Runlist commands.
    """
    pass


@tools.group(name='crashlog')
def crashlog_group():
    """
    Crashlog commands.
    """
    pass


@tools.group(name='group')
def group_group():
    """
    Routing group commands.
    """
    pass


@tools.group(name='tracing')
def tracing_group():
    """
    Dynamic tracing support.
    """
    pass


@tools.group(name='logging')
def logging_group():
    """
    Dynamic logging filtering support.
    """
    pass


@tools.group(name='timeouts')
def timeouts_group():
    """
    Configurable timeout for applications support.
    """
    pass


@tools.group(name='auth')
def auth_group():
    """
    Authorization tokens management.
    """
    _print_experimental_warning()


@tools.group(name='access')
def access_group():
    """
    ACL management.
    """
    _print_experimental_warning()


@access_group.group(name='storage')
def access_storage_group():
    """
    ACL management for storage collections.
    """
    pass


@access_group.group(name='event')
def access_event_group():
    """
    ACL management for services.
    """
    pass


@tools.group(name='keyring')
def keyring_group():
    """
    Public keys management.
    """
    pass


@tools.command()
@click.option('--type', 'ty', default='plain', type=Choice(['plain', 'json']), help='Output type.')
@click.option('--query', help='Filtering query.')
@click.option('--query-type', default='mql', type=click.Choice(['mql', 'ast']), help='Query type.')
@with_options
def metrics(ty, query, query_type, **kwargs):
    """
    Outputs runtime metrics collected from cocaine-runtime and its services.

    This command shows runtime metrics collected from cocaine-runtime and its services during their
    lifetime.
    There are four kind of metrics available: gauges, counters, meters and timers.

    \b
    - Gauges   - an instantaneous measurement of a value.
    - Counters - just a gauge for an atomic integer instance.
    - Meters   - measures the rate of events over time (e.g., "requests per second"). In addition
      to the mean rate, meters also track 1-, 5-, and 15-minute moving averages.
    - Timers   - measures both the rate that a particular piece of code is called and the
      distribution of its duration.


    Every metric in has a unique name, which is just a dotted-name string like "connections.count"
    or "node.queue.size".

    An output type can be configured using --type option. The default one results in plain
    formatting where there is only one depth level.

    As an alternative you can expanded the JSON tree by specifying --type=json option. The depth of
    the result tree depends on metric name which is split by dot symbol.

    The result output will be probably too large without any customization. To reduce this output
    there are custom filters, which can be specified using --query option. Technically it's a
    special metrics query language (MQL) which supports the following operations and functions:

    \b
    - contains(<expr>, <expr>) - checks whether the result of second expression contains in the
      result of first expression. These expressions must resolve in strings. An output type of this
      function is bool.
    - name() - resolves in metric name.
    - type() - resolves in metric type (counter, meter, etc.).
    - tag(<expr>) - extracts custom metric tag and results in string.
    - && - combines several expressions in one, which applies when all of them apply.
    - || - combines several expressions in one, which applies when any of them apply.
    - == - compares two expressions for equality.
    - != - compares two expressions for an non-equality.
    - Also string literals (alphanumeric with dots) can be used as an expressions, for
      example "name() == locator.connections.accepted".

    Priorities can be specified using braces as in usual math expressions.

    The grammar for this query language is:

    \b
    expr    ::= term ((AND | OR) term)*
    term    ::= factor ((EQ | NE) factor)*
    factor  ::= func | literal | number | LPAREN expr RPAREN
    func    ::= literal LPAREN expr (,expr)* RPAREN
    literal ::= alphanum | .
    number  ::= <floating point number>

    An example of the query, which returns all meters (for all services) and the number of accepted
    connections for the Locator
    service: "contains(type(), meter) || name() == locator.connections.accepted".
    """
    ctx = Context(**kwargs)
    ctx.execute_action('metrics', **{
        'metrics': ctx.repo.create_secure_service('metrics'),
        'ty': ty,
        'query': query,
        'query_type': query_type,
    })


@app_group.command(name='list')
@with_options
def app_list(**kwargs):
    """
    Show uploaded applications.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('app:list', **{
        'storage': ctx.repo.create_secure_service('storage'),
    })


@app_group.command(name='view')
@click.option('-n', '--name', metavar='', help='Application name.')
@with_options
def app_view(name, **kwargs):
    """
    Show manifest content for an application.

    If application is not uploaded, an error will be displayed.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('app:view', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
    })


@app_group.command(name='upload')
@click.argument('path', required=False, type=click.Path(exists=True))
@click.option('-n', '--name', metavar='', help='Application name.')
@click.option('--manifest', metavar='', help='Manifest file name.')
@click.option('--package', metavar='', help='Path to the application archive.')
@click.option('--docker_address', metavar='', help='Docker address.')
@click.option('--registry', metavar='', help='Docker Registry address.')
@click.option('--manifest-only', is_flag=True, metavar='', default=False, help='Upload only manifest.')
@with_options
def app_upload(path, name, manifest, package, docker_address, registry, manifest_only, **kwargs):
    """
    Upload application with its environment (directory) into the storage.

    Application directory or its subdirectories must contain valid manifest file
    named `manifest.json` or `manifest` otherwise you must specify it explicitly by
    setting `--manifest` option.

    You can specify application name. By default, leaf directory name is treated as application
    name.

    If you have already prepared application archive (*.tar.gz), you can explicitly specify path to
    it by setting `--package` option.

    Additional output can be turned on by passing `-vvvv` option.
    """
    lower_limit = 120.0

    ctx = Context(**kwargs)
    if ctx.timeout < lower_limit:
        ctx.timeout = lower_limit
        log.info('shifted timeout to the %.2fs', ctx.timeout)

    mutex_record = collections.namedtuple('mutex_record', 'value, name')
    mutex = [
        (mutex_record(path, 'PATH'), mutex_record(package, '--package')),
        (mutex_record(package, '--package'), mutex_record(docker_address, '--docker')),
        (mutex_record(package, '--package'), mutex_record(registry, '--registry')),
    ]
    for (f, s) in mutex:
        if f.value and s.value:
            click.echo('Wrong usage: option {} and {} are mutual exclusive, you can only use one'.
                       format(f.name, s.name))
            exit(os.EX_USAGE)

    if manifest_only:
        ctx.execute_action('app:upload-manual', **{
            'storage': ctx.repo.create_secure_service('storage'),
            'name': name,
            'manifest': manifest,
            'package': None,
            'manifest_only': manifest_only,
        })
    elif package:
        ctx.execute_action('app:upload-manual', **{
            'storage': ctx.repo.create_secure_service('storage'),
            'name': name,
            'manifest': manifest,
            'package': package
        })
    elif docker_address:
        ctx.execute_action('app:upload-docker', **{
            'storage': ctx.repo.create_secure_service('storage'),
            'path': path,
            'name': name,
            'manifest': manifest,
            'address': docker_address,
            'registry': registry
        })
    else:
        ctx.execute_action('app:upload', **{
            'storage': ctx.repo.create_secure_service('storage'),
            'path': path,
            'name': name,
            'manifest': manifest
        })


@app_group.command(name='import')
@click.argument('path', type=click.Path(exists=True))
@click.option('-n', '--name', metavar='', help='Application name.')
@click.option('--manifest', metavar='', help='Manifest file name.')
@click.option('--container_url', metavar='', required=True, help='Docker container url.')
@click.option('--docker_address', metavar='', required=True, help='Docker address.')
@click.option('--registry', metavar='', help='Docker Registry address.')
@with_options
def app_import(path, name, manifest, container_url, docker_address, registry, **kwargs):
    """
    Import application Docker container.
    """
    lower_limit = 120.0

    ctx = Context(**kwargs)
    if ctx.timeout < lower_limit:
        ctx.timeout = lower_limit
        log.info('shifted timeout to the %.2fs', ctx.timeout)

    if container_url and docker_address:
        ctx.execute_action('app:import-docker', **{
            'storage': ctx.repo.create_secure_service('storage'),
            'path': path,
            'name': name,
            'manifest': manifest,
            'container': container_url,
            'address': docker_address,
            'registry': registry
        })
    else:
        raise ValueError("both `container_url` and `docker_address` options must not be empty")


@app_group.command(name='remove')
@click.option('-n', '--name', metavar='', help='Application name.')
@with_options
def app_remove(name, **kwargs):
    """
    Remove application from storage.

    No error messages will display if specified application is not uploaded.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('app:remove', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
    })


@app_group.command(name='start')
@click.option('-n', '--name', metavar='', help='Application name.')
@click.option('-r', '--profile', metavar='', help='Profile name.')
@with_options
def app_start(name, profile, **kwargs):
    """
    Start an application with specified profile.

    Does nothing if application is already running.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('app:start', **{
        'node': ctx.repo.create_secure_service('node'),
        'name': name,
        'profile': profile
    })


@app_group.command(name='pause')
@click.option('-n', '--name', metavar='', help='Application name.')
@with_options
def app_pause(name, **kwargs):
    """
    Stop application.

    This command is alias for ```cocaine-tool app stop```.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('app:stop', **{
        'node': ctx.repo.create_secure_service('node'),
        'name': name,
    })


@app_group.command(name='stop')
@click.option('-n', '--name', metavar='', help='Application name.')
@with_options
def app_stop(name, **kwargs):
    """
    Stop application.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('app:stop', **{
        'node': ctx.repo.create_secure_service('node'),
        'name': name,
    })


@app_group.command(name='restart')
@click.option('-n', '--name', metavar='', help='Application name.')
@click.option('-r', '--profile', metavar='', help='Profile name.')
@with_options
def app_restart(name, profile, **kwargs):
    """
    Restart application.

    Executes ```cocaine-tool app pause``` and ```cocaine-tool app start``` sequentially.

    It can be used to quickly change application profile.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('app:restart', **{
        'node': ctx.repo.create_secure_service('node'),
        'locator': ctx.locator,
        'name': name,
        'profile': profile,
    })


@app_group.command()
@click.option('-n', '--name', metavar='', help='Application name.')
@with_options
def check(name, **kwargs):
    """
    Check application status.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('app:check', **{
        'node': ctx.repo.create_secure_service('node'),
        'name': name,
    })


@profile_group.command(name='list')
@with_options
def profile_list(**kwargs):
    """
    Show uploaded profiles.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('profile:list', **{
        'storage': ctx.repo.create_secure_service('storage'),
    })


@profile_group.command(name='view')
@click.option('-n', '--name', metavar='', required=True, help='Profile name.')
@with_options
def profile_view(name, **kwargs):
    """
    Show profile configuration content.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('profile:view', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
    })


@profile_group.command(name='upload')
@click.option('-n', '--name', metavar='', required=True, help='Profile name.')
@click.option('-r', '--profile', metavar='', required=True, help='Path to profile.')
@with_options
def profile_upload(name, profile, **kwargs):
    """
    Upload profile into the storage.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('profile:upload', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
        'profile': profile,
    })


@profile_group.command(name='edit')
@click.option('-n', '--name', metavar='', required=True, help='Profile name.')
@with_options
def profile_edit(name, **kwargs):
    """
    Edit profile using interactive editor.
    """
    ctx = Context(**kwargs)
    ctx.timeout = None
    ctx.execute_action('profile:edit', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
    })


@profile_group.command(name='remove')
@click.option('-n', '--name', metavar='', required=True, help='Profile name.')
@with_options
def profile_remove(name, **kwargs):
    """
    Remove profile from the storage.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('profile:remove', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
    })


@profile_group.command(name='copy')
@click.option('-n', '--name', metavar='', required=True, help='Profile name.')
@click.option('--copyname', metavar='', required=True, help='Profile new name.')
@with_options
def profile_copy(name, copyname, **kwargs):
    """
    Copy a profile.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('profile:copy', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
        'copyname': copyname,
    })


@profile_group.command(name='rename')
@click.option('-n', '--name', metavar='', required=True, help='Profile name.')
@click.option('--copyname', metavar='', required=True, help='Profile new name.')
@with_options
def profile_rename(name, copyname, **kwargs):
    """
    Rename a profile.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('profile:rename', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
        'copyname': copyname,
    })


@runlist_group.command(name='list')
@with_options
def runlist_list(**kwargs):
    """
    Show uploaded runlists.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('runlist:list', **{
        'storage': ctx.repo.create_secure_service('storage'),
    })


@runlist_group.command(name='view')
@click.option('-n', '--name', metavar='', required=True, help='Runlist name.')
@with_options
def runlist_view(name, **kwargs):
    """
    Show configuration content for a specified runlist.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('runlist:view', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name
    })


@runlist_group.command(name='edit')
@click.option('-n', '--name', metavar='', required=True, help='Runlist name.')
@with_options
def runlist_edit(name, **kwargs):
    """
    Edit specified runlist interactively using editor.
    """
    ctx = Context(**kwargs)
    ctx.timeout = None
    ctx.execute_action('runlist:edit', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name
    })


@runlist_group.command(name='upload')
@click.option('-n', '--name', metavar='', required=True, help='Runlist name.')
@click.option('--runlist', metavar='', required=True,
              help='Either path to the runlist file or inline content')
@with_options
def runlist_upload(name, runlist, **kwargs):
    """
    Upload runlist with context into the storage.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('runlist:upload', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
        'runlist': runlist,
    })


@runlist_group.command(name='create')
@click.option('-n', '--name', metavar='', required=True, help='Runlist name.')
@with_options
def runlist_create(name, **kwargs):
    """
    Create runlist and upload it into the storage.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('runlist:create', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
    })


@runlist_group.command(name='remove')
@click.option('-n', '--name', metavar='', required=True, help='Runlist name.')
@with_options
def runlist_remove(name, **kwargs):
    """
    Remove runlist from the storage.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('runlist:remove', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
    })


@runlist_group.command(name='copy')
@click.option('-n', '--name', metavar='', required=True, help='Runlist name.')
@click.option('-c', '--copyname', metavar='', required=True, help='Cloned runlist name.')
@with_options
def runlist_copy(name, copyname, **kwargs):
    """
    Clone runlist.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('runlist:copy', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
        'copyname': copyname,
    })


@runlist_group.command(name='rename')
@click.option('-n', '--name', metavar='', required=True, help='Runlist name.')
@click.option('-c', '--copyname', metavar='', required=True, help='Runlist new name.')
@with_options
def runlist_rename(name, copyname, **kwargs):
    """
    Rename runlist.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('runlist:rename', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
        'copyname': copyname,
    })


@runlist_group.command(name='add-app')
@click.option('-n', '--name', metavar='', required=True, help='Runlist name.')
@click.option('--app', metavar='', required=True, help='Application name.')
@click.option('--profile', metavar='', required=True, help='Suggested profile name.')
@click.option('-f', '--force', is_flag=True, default=False, help='Create runlist if not exists.')
@with_options
def runlist_add_app(name, app, profile, force, **kwargs):
    """
    Add specified application with profile to the specified runlist.

    Existence of application or profile is not checked.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('runlist:add-app', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
        'app': app,
        'profile': profile,
        'force': force
    })


@runlist_group.command(name='remove-app')
@click.option('-n', '--name', metavar='', required=True, help='Runlist name.')
@click.option('--app', metavar='', required=True, help='Application name.')
@with_options
def runlist_remove_app(name, app, **kwargs):
    """
    Remove specified application from the runlist.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('runlist:remove-app', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
        'app': app,
    })


@crashlog_group.command(name='status')
@with_options
def crashlog_status(**kwargs):
    """
    Show crashlogs status.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('crashlog:status', **{
        'storage': ctx.repo.create_secure_service('storage'),
    })


@crashlog_group.command(name='list')
@click.option('-n', '--name', metavar='', required=True, help='Crashlog name.')
@click.option('-d', '--day', metavar='', help='Date filter.')
@with_options
def crashlog_list(name, day, **kwargs):
    """
    Show crashlogs list for application.

    Prints crashlog list in "Unix Timestamp - UUID" format.

    Filtering by day accepts the following variants: today, yesterday, %d, %d-%m or %d-%m-%Y.
    For example given today is a 06-12-2016 crashlogs for yesterday can be listed using:
    `--day=yesterday`, `--day=5`, `--day=05-12` or `--day=05-12-2016`.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('crashlog:list', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
        'day_string': day,
    })


@crashlog_group.command(name='view')
@click.option('-n', '--name', metavar='', required=True, help='Crashlog name.')
@click.option('-t', '--timestamp', metavar='', help='Timestamp.')
@with_options
def crashlog_view(name, timestamp, **kwargs):
    """
    Show crashlog for application with specified timestamp.

    Last crashlog for a given application will be displayed unless timestamp option is specified.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('crashlog:view', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
        'timestamp': timestamp,
    })


@crashlog_group.command(name='remove')
@click.option('-n', '--name', metavar='', required=True, help='Crashlog name.')
@click.option('-t', '--timestamp', metavar='', help='Timestamp.')
@with_options
def crashlog_remove(name, timestamp, **kwargs):
    """
    Remove crashlog for application with specified timestamp from the storage.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('crashlog:remove', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
        'timestamp': timestamp,
    })


@crashlog_group.command(name='removeall')
@click.option('-n', '--name', metavar='', required=True, help='Crashlog name.')
@with_options
def crashlog_removeall(name, **kwargs):
    """
    Remove all crashlogs for application from the storage.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('crashlog:removeall', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
    })


@crashlog_group.command(name='clean')
@click.option('-n', '--name', metavar='', required=True, help='Crashlog name.')
@click.option('-t', '--timestamp', metavar='', help='Timestamp.')
@click.option('-s', '--size', metavar='', default=1000, help='Number of crashlogs to leave.')
@with_options
def crashlog_clean(name, timestamp, size, **kwargs):
    """
    For application NAME leave SIZE crashlogs or remove all crashlogs with timestamp > TIMESTAMP.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('crashlog:clean', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
        'size': size,
        'timestamp': timestamp,
    })


@crashlog_group.command(name='cleanrange')
@click.option('--from_day', metavar='', required=True, help='From day.')
@click.option('--up_to_day', metavar='', default='yesterday', help='Up to day.')
@with_options
def crashlog_cleanrange(from_day, up_to_day, **kwargs):
    """
    Remove all crashlogs from one date up to another.

    The date can be specified as DAY-[MONTH-[YEAR]].

    Example:
        today, yesterday, 10, 10-09, 10-09-2015
    """
    ctx = Context(**kwargs)
    ctx.execute_action('crashlog:cleanwhen', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'from_day': from_day,
        'to_day': up_to_day,
    })


@group_group.command(name='list')
@with_options
def group_list(**kwargs):
    """
    Show available routing groups.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('group:list', **{
        'storage': ctx.repo.create_secure_service('storage'),
    })


@group_group.command(name='view')
@click.option('-n', '--name', metavar='', required=True, help='Routing group name.')
@with_options
def group_view(name, **kwargs):
    """
    Show specified routing group.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('group:view', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
    })


@group_group.command(name='create')
@click.option('-n', '--name', metavar='', required=True, help='Routing group name.')
@click.option('-c', '--content', metavar='', required=True, help='Routing group content.')
@with_options
def group_create(name, content, **kwargs):
    """
    Create routing group.

    You can optionally specify content for created routing group. It can be either direct JSON
    expression in single quotes, or path to the json file with settings. Settings itself must be
    key-value list, where `key` represents application name, and `value` represents its weight.

    For example:

    cocaine-tool group create -n new_group -c '{
        "app": 1,
        "another_app": 2
    }'.

    Warning: all application weights must be positive integers, total weight must be positive.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('group:create', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
        'content': content,
    })


@group_group.command(name='remove')
@click.option('-n', '--name', metavar='', required=True, help='Routing group name.')
@with_options
def group_remove(name, **kwargs):
    """
    Remove routing group from the storage.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('group:remove', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
    })


@group_group.command(name='copy')
@click.option('-n', '--name', metavar='', required=True, help='Routing group name.')
@click.option('-c', '--copyname', metavar='', required=True, help='Cloned routing group name.')
@with_options
def group_copy(name, copyname, **kwargs):
    """
    Copy routing group.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('group:copy', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
        'copyname': copyname,
    })


@group_group.command(name='rename')
@click.option('-n', '--name', metavar='', required=True, help='Routing group name.')
@click.option('-c', '--copyname', metavar='', required=True, help='New routing group name.')
@with_options
def group_rename(name, copyname, **kwargs):
    """
    Rename routing group.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('group:rename', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
        'copyname': copyname,
    })


@group_group.command(name='refresh')
@click.option('-n', '--name', metavar='', help='Routing group name.')
@with_options
def group_refresh(name, **kwargs):
    """
    Refresh routing group.

    If the name option is empty, this command will refresh all groups.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('group:refresh', **{
        'locator': ctx.locator,
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
    })


@group_group.command(name='push')
@click.option('-n', '--name', metavar='', required=True, help='Routing group name.')
@click.option('--app', metavar='', required=True, help='Application name.')
@click.option('-w', '--weight', metavar='', required=True, type=int, help='Application weight.')
@with_options
def group_push(name, app, weight, **kwargs):
    """
    Add application with its weight into the routing group.

    Warning: application weight must be positive integer.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('group:app:add', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
        'app': app,
        'weight': weight,
    })


@group_group.command(name='pop')
@click.option('-n', '--name', metavar='', required=True, help='Routing group name.')
@click.option('--app', metavar='', required=True, help='Application name.')
@with_options
def group_pop(name, app, **kwargs):
    """
    Remove application from the specified routing group.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('group:app:remove', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
        'app': app,
    })


@group_group.command(name='edit')
@click.option('-n', '--name', metavar='', required=True, help='Routing group name.')
@with_options
def group_edit(name, **kwargs):
    """
    Edit specified routing group in an interactive editor.
    """
    ctx = Context(**kwargs)
    ctx.timeout = None
    ctx.execute_action('group:edit', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
    })


@tracing_group.command(name='store')
@click.option('-n', '--name', metavar='', required=True, help='Unicorn node.')
@click.option('-v', '--value', metavar='', type=float, required=True, help='Value in percents.')
@with_options
def tracing_store(name, value, **kwargs):
    ctx = Context(**kwargs)
    ctx.execute_action('tracing:store', **{
        'configuration_service': ctx.repo.create_secure_service('unicorn'),
        'name': name,
        'value': value,
    })


@tracing_group.command(name='remove')
@click.option('-n', '--name', metavar='', required=True, help='Unicorn node.')
@with_options
def tracing_remove(name, **kwargs):
    ctx = Context(**kwargs)
    ctx.execute_action('tracing:remove', **{
        'configuration_service': ctx.repo.create_secure_service('unicorn'),
        'name': name,
    })


@tracing_group.command(name='view')
@click.option('-n', '--name', metavar='', help='Unicorn node.')
@with_options
def tracing_view(name, **kwargs):
    ctx = Context(**kwargs)
    ctx.execute_action('tracing:view', **{
        'configuration_service': ctx.repo.create_secure_service('unicorn'),
        'name': name,
    })


@timeouts_group.command(name='store')
@click.option('-n', '--name', metavar='', required=True, help='Unicorn node.')
@click.option('-e', '--event', metavar='', required=True, help='Event name.')
@click.option('-v', '--value', metavar='', required=True, default=30.0, help='Seconds.')
@with_options
def timeouts_store(name, event, value, **kwargs):
    ctx = Context(**kwargs)
    ctx.execute_action('timeouts:store', **{
        'configuration_service': ctx.repo.create_secure_service('unicorn'),
        'name': name,
        'value': value,
        'event': event,
    })


@timeouts_group.command(name='remove')
@click.option('-n', '--name', metavar='', required=True, help='Unicorn node.')
@click.option('-e', '--event', metavar='', required=True, help='Event name.')
@with_options
def timeouts_remove(name, event, **kwargs):
    ctx = Context(**kwargs)
    ctx.execute_action('timeouts:remove', **{
        'configuration_service': ctx.repo.create_secure_service('unicorn'),
        'name': name,
        'event': event,
    })


@timeouts_group.command(name='view')
@click.option('-n', '--name', metavar='', required=True, help='Unicorn node.')
@with_options
def timeouts_view(name, **kwargs):
    ctx = Context(**kwargs)
    ctx.execute_action('timeouts:view', **{
        'configuration_service': ctx.repo.create_secure_service('unicorn'),
        'name': name,
    })


@timeouts_group.command(name='drop')
@click.option('-n', '--name', metavar='', required=True, help='Unicorn node.')
@with_options
def timeouts_drop(name, **kwargs):
    ctx = Context(**kwargs)
    ctx.execute_action('timeouts:drop', **{
        'configuration_service': ctx.repo.create_secure_service('unicorn'),
        'name': name,
    })


@logging_group.command(name='list_loggers')
@with_options
def logging_list_loggers(**kwargs):
    """
    List all registered logger names.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('logging:list_loggers', **{
        'logging_service': ctx.repo.create_secure_service('logging'),
    })


@logging_group.command(name='set_filter')
@click.option('-n', '--name', metavar='', required=True, help='Logger name.')
@click.option('--filter_def', metavar='', required=True, help='Filter definition.')
@click.option('--ttl', metavar='', required=True, help='TTL.')
@with_options
def logging_set_filter(name, filter_def, ttl, **kwargs):
    """
    Set local filter.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('logging:set_filter', **{
        'logging_service': ctx.repo.create_secure_service('logging'),
        'logger_name': name,
        'filter_def': filter_def,
        'ttl': ttl,
    })


@logging_group.command(name='remove_filter')
@click.option('-i', '--filter-id', metavar='', required=True, help='Filter id.')
@with_options
def logging_remove_filter(filter_id, **kwargs):
    """
    Remove filter by filter id.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('logging:remove_filter', **{
        'logging_service': ctx.repo.create_secure_service('logging'),
        'filter_id': filter_id,
    })


@logging_group.command(name='list_filters')
@with_options
def logging_list_filters(**kwargs):
    """
    List all available filters.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('logging:list_filters', **{
        'logging_service': ctx.repo.create_secure_service('logging'),
    })


@logging_group.command(name='set_cluster_filter')
@click.option('-n', '--name', metavar='', required=True, help='Logger name.')
@click.option('--filter_def', metavar='', required=True, help='Filter definition.')
@click.option('--ttl', metavar='', required=True, help='TTL.')
@with_options
def logging_set_cluster_filter(name, filter_def, ttl, **kwargs):
    """
    Set cluster-wide filter.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('logging:set_cluster_filter', **{
        'logging_service': ctx.repo.create_secure_service('logging'),
        'logger_name': name,
        'filter_def': filter_def,
        'ttl': ttl,
    })


@auth_group.command(name='list')
@with_options
def auth_list(**kwargs):
    """
    Shows available authorization groups.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('auth:group:list', **{
        'storage': ctx.repo.create_secure_service('storage'),
    })


@auth_group.command(name='create')
@click.option('-n', '--name', metavar='', required=True, help='Group name.')
@click.option('--token', metavar='', required=True, help='Secure token.')
@click.option('--force', metavar='', is_flag=True, default=False, help='Override if exists.')
@with_options
def auth_create(name, token, force, **kwargs):
    """
    Creates an authorization group.

    The group sets a named association between an authorization token and the list of services. This
    is useful for group of applications that want to share a single token.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('auth:group:create', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
        'token': token,
        'force': force,
    })


@auth_group.command(name='edit')
@click.option('-n', '--name', metavar='', required=True, help='Group name.')
@with_options
def auth_edit(name, **kwargs):
    """
    Interactively edits an authorization group.
    """
    ctx = Context(**kwargs)
    ctx.timeout = None
    ctx.execute_action('auth:group:edit', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
    })


@auth_group.command(name='rm')
@click.option('-n', '--name', metavar='', required=True, help='Group name.')
@click.option('--drop/--no-drop', metavar='', default=False, help='Remove members.')
@with_options
def auth_remove(name, drop, **kwargs):
    """
    Removes an authorization group.

    Removes an authorization group with or without excluding associated members depending on --drop
    flag (disabled by default).
    """
    ctx = Context(**kwargs)
    ctx.execute_action('auth:group:remove', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
        'drop': drop,
    })


@auth_group.command(name='view')
@click.option('-n', '--name', metavar='', required=True, help='Group name.')
@with_options
def auth_view(name, **kwargs):
    """
    Shows an authorization group's content.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('auth:group:view', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
    })


@auth_group.command(name='add')
@click.option('-n', '--name', metavar='', required=True, help='Group name.')
@click.option('--service', metavar='', required=True, help='Service name.')
@with_options
def auth_add(name, service, **kwargs):
    """
    Adds a member of an authorization group.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('auth:group:members:add', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
        'service': service,
    })


@auth_group.command(name='exclude')
@click.option('-n', '--name', metavar='', required=True, help='Group name.')
@click.option('--service', metavar='', required=True, help='Service name.')
@with_options
def auth_exclude(name, service, **kwargs):
    """
    Excludes a member of an authorization group.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('auth:group:members:exclude', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
        'service': service,
    })


@access_storage_group.command(name='list')
@with_options
def access_storage_list(**kwargs):
    """
    Shows collections with ACL.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('access:storage:list', **{
        'storage': ctx.repo.create_secure_service('storage'),
    })


@access_storage_group.command(name='view')
@click.option('-n', '--name', metavar='', required=True, help='Collection name.')
@with_options
def access_storage_view(name, **kwargs):
    """
    Shows ACL for the specified collection.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('access:storage:view', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
    })


@access_storage_group.command(name='create')
@click.option('-n', '--name', metavar='', required=True, help='Collection name.')
@with_options
def access_storage_create(name, **kwargs):
    """
    Creates new ACL for the specified collection.

    Does nothing if ACL already exists.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('access:storage:create', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
    })


@access_storage_group.command(name='edit')
@click.option('-n', '--name', metavar='', required=True, help='Collection name.')
@click.option('-c', '--cid', metavar='', multiple=True, help='Client ID.')
@click.option('-u', '--uid', metavar='', multiple=True, help='User ID.')
@click.option('--perm', metavar='', required=True, type=click.Choice(['R', 'W', 'RW', '0']),
              help='Permissions.')
@with_options
def access_storage_edit(name, cid, uid, perm, **kwargs):
    """
    Edits ACL for the specified collection.

    Creates if necessary.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('access:storage:edit', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
        'cids': cid,
        'uids': uid,
        'perm': perm,
    })


@access_storage_group.command(name='rm')
@click.option('-n', '--name', metavar='', help='Collection name.')
@click.option('--yes', is_flag=True, default=False, help='Do not prompt.')
@with_options
def access_storage_rm(name, yes, **kwargs):
    """
    Remove ACL for the specified collection.

    If none is specified - removes ACL for all collections.
    """
    if name is None:
        if not yes:
            click.confirm('Are you sure you want to remove all ACL?', abort=True)

    ctx = Context(**kwargs)
    ctx.execute_action('access:storage:rm', **{
        'storage': ctx.repo.create_secure_service('storage'),
        'name': name,
    })


@access_event_group.command(name='list')
@with_options
def access_list(**kwargs):
    """
    Shows services for which there are ACL specified.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('access:list', **{
        'unicorn': ctx.repo.create_secure_service('unicorn'),
    })


@access_event_group.command(name='view')
@click.option('--name', metavar='', required=True, help='Service name.')
@with_options
def access_view(name, **kwargs):
    """
    Shows ACL for the specified service.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('access:view', **{
        'unicorn': ctx.repo.create_secure_service('unicorn'),
        'service': name,
    })


@access_event_group.command(name='add')
@click.option('--name', metavar='', required=True, help='Service name.')
@click.option('--event', metavar='', required=True, help='Event name.')
@click.option('-c', '--cid', multiple=True, help='Client identifier.')
@click.option('-u', '--uid', multiple=True, help='User identifier.')
@with_options
def access_add(name, event, cid, uid, **kwargs):
    """
    Creates a new record with specified cid/uid in the event authorization.

    Requests with token that contains such cid/uid will have access to the specified event of a
    service.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('access:add', **{
        'unicorn': ctx.repo.create_secure_service('unicorn'),
        'service': name,
        'event': event,
        'cids': cid,
        'uids': uid,
    })


@access_event_group.command(name='edit')
@click.option('--name', metavar='', required=True, help='Service name.')
@with_options
def access_edit(name, **kwargs):
    """
    Edits interactively an access list for a service with further validation.
    """
    ctx = Context(**kwargs)
    ctx.timeout = None
    ctx.execute_action('access:edit', **{
        'unicorn': ctx.repo.create_secure_service('unicorn'),
        'service': name,
    })


@keyring_group.command(name='view')
@with_options
def keyring_view(**kwargs):
    """
    View saved public keys.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('keyring:view', **{
        'storage': ctx.repo.create_secure_service('storage'),
    })


@keyring_group.command(name='update')
@click.option('--cid', metavar='', type=int, required=True, help='Client id.')
@with_options
def keyring_update(cid, **kwargs):
    """
    Downloads and saves public key(s).

    Downloads public key(s) from the TVM service for a given client id and saves in the storage
    for further usage.
    Note, that the cocaine-runtime automatically refreshes its keyring after this call.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('keyring:update', **{
        'tvm': ctx.repo.create_secure_service('tvm'),
        'cid': cid,
    })


@keyring_group.command(name='remove')
@click.option('-k', '--key', metavar='', help='Public key')
@click.option('--yes', is_flag=True, default=False, help='Do not prompt.')
@with_options
def keyring_remove(key, yes, **kwargs):
    """
    Removes a public key from the keyring.

    Does nothing if a key is already not in the keyring. If none is specified - clears the keyring.
    To force the cocaine-runtime to refresh its keyring, call `refresh` method.
    """
    if key is None:
        if not yes:
            click.confirm('Are you sure you want to remove all keys?', abort=True)

    ctx = Context(**kwargs)
    ctx.execute_action('keyring:remove', **{
        'key': key,
        'storage': ctx.repo.create_secure_service('storage'),
    })


@keyring_group.command(name='edit')
@with_options
def keyring_edit(**kwargs):
    """
    Edits interactively the keyring.
    """
    ctx = Context(**kwargs)
    ctx.timeout = None
    ctx.execute_action('keyring:edit', **{
        'storage': ctx.repo.create_secure_service('storage'),
    })


@keyring_group.command(name='refresh')
@with_options
def keyring_refresh(**kwargs):
    """
    Refresh the keyring in the cocaine-runtime.
    """
    ctx = Context(**kwargs)
    ctx.execute_action('keyring:refresh', **{
        'tvm': ctx.repo.create_secure_service('tvm'),
    })


cli = click.CommandCollection(sources=[tools])



#
# Copyright (c) 2013+ Anton Tyurin <noxiouz@yandex.ru>
# Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
# Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
# This file is part of Cocaine-tools.
#
# Cocaine is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Cocaine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from cocaine.decorators import coroutine
from cocaine.exceptions import ServiceError

from cocaine.tools.error import ToolsError
from cocaine.tools import actions
from cocaine.tools.actions import CocaineConfigReader, log
from cocaine.tools.printer import printer
from cocaine.tools.tags import PROFILES_TAGS

__author__ = 'Evgeny Safronov <division494@gmail.com>'


@coroutine
def upload_profile(storage, name, profile):
    try:
        channel = yield storage.write('profiles', name, profile, PROFILES_TAGS)
        yield channel.rx.get()
    except ServiceError as err:
        error_message = 'unable to write profile "%s" to storage: %s' % (name, err)
        log.error(error_message)
        raise ToolsError(error_message)


class Specific(actions.Specific):
    def __init__(self, storage, name):
        super(Specific, self).__init__(storage, 'profile', name)


class List(actions.List):
    def __init__(self, storage):
        super(List, self).__init__('profiles', PROFILES_TAGS, storage)


class View(actions.View):
    def __init__(self, storage, name):
        super(View, self).__init__(storage, 'profile', name, 'profiles')


class Upload(Specific):
    def __init__(self, storage, name, profile):
        super(Upload, self).__init__(storage, name)
        self.profile = profile
        if isinstance(self.profile, dict):
            return
        elif isinstance(self.profile, (str, unicode)) and len(self.profile.strip()) > 0:
            return
        if not self.profile:
            raise ValueError('Please specify profile file path')

    @coroutine
    def execute(self):
        with printer('Loading profile'):
            profile = CocaineConfigReader.load(self.profile)
        with printer('Uploading "%s"', self.name):
            yield upload_profile(self.storage, self.name, profile)


class Remove(Specific):
    @coroutine
    def execute(self):
        log.info('Removing "%s"... ', self.name)
        channel = yield self.storage.remove('profiles', self.name)
        yield channel.rx.get()
        log.info('OK')


class Copy(Specific):
    def __init__(self, storage, name, copyname):
        super(Copy, self).__init__(storage, name)
        self.copyname = copyname
        if self.name == self.copyname:
            raise ToolsError("unable to copy an instance to itself")

    @coroutine
    def execute(self):
        log.info('Rename "%s" to "%s"', self.name, self.copyname)
        oldprofile = yield View(self.storage, self.name).execute()
        yield Upload(self.storage, self.copyname, oldprofile).execute()
        log.info('OK')


class Rename(Copy):
    @coroutine
    def execute(self):
        yield super(Rename, self).execute()
        yield Remove(self.storage, self.name).execute()



#
# Copyright (c) 2013+ Anton Tyurin <noxiouz@yandex.ru>
# Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
# Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
# This file is part of Cocaine-tools.
#
# Cocaine is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Cocaine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import json
import logging
import os
from signal import SIGTERM
import time

from cocaine.proxy import Daemon
from cocaine.proxy.proxy import CocaineProxy
from cocaine.tools import log


class Error(Exception):
    pass


class Start(object):
    def __init__(self, port, count, config, daemon, pidfile):
        self.port = port
        self.count = count
        self.config = config
        self.daemon = daemon
        self.pidfile = pidfile

    def execute(self):
        log.info('Starting cocaine proxy... ')
        config = self.loadConfig()

        if self.daemon:
            self.checkPermissions()
            daemon = Daemon(self.pidfile)
            daemon.run = self.run
            daemon.start(config)
        else:
            log.info('OK')
            self.run(config)

    def checkPermissions(self):
        if os.path.exists(self.pidfile):
            log.error('FAIL')
            raise Error('is already running (pid file "{0}" exists)'.format(self.pidfile))
        else:
            try:
                with open(self.pidfile, 'w'):
                    pass
                os.remove(self.pidfile)
                log.info('OK')
            except IOError as err:
                log.error('FAIL')
                raise Error('failed to create pid file - {0}'.format(err))

    def loadConfig(self):
        config = {
            'cache': 5
        }
        try:
            with open(self.config, 'r') as fh:
                config = json.loads(fh.read())
        except IOError as err:
            log.error('Failed to load config - %s. Default config will be used.', err)
        except Exception as err:
            log.error('Unexpected error occurs while loading config - %s. Default config will be used.', err)
        finally:
            return config

    def run(self, config):
        if 'logging' in config:
            logging.config.dictConfig(config['logging'])
        else:
            self.configureLogging()

        proxy = CocaineProxy(self.port, **config)
        proxy.run(self.count)

    @staticmethod
    def configureLogging():
        formatter = logging.Formatter('[%(asctime)s] %(levelname)-8s: %(message)s')
        handler = logging.FileHandler('/var/log/cocaine-python-proxy.log')
        handler.setFormatter(formatter)
        handler.setLevel(logging.INFO)
        proxyLog = logging.getLogger('cocaine.proxy')
        proxyLog.addHandler(handler)
        proxyLog.setLevel(logging.INFO)


class Stop(object):
    def __init__(self, pidfile):
        self.pidfile = pidfile

    def execute(self):
        log.info('Stopping cocaine proxy... ')
        try:
            with open(self.pidfile, 'r') as fh:
                pid = int(fh.read().strip())
        except IOError:
            pid = None

        if not pid:
            log.error('FAIL')
            log.error('Cocaine proxy is not running')
            exit(1)

        try:
            elapsed = 0
            while True and elapsed < 30.0:
                os.kill(pid, SIGTERM)
                time.sleep(0.5)
                elapsed += 0.5

            if elapsed > 30.0:
                os.kill(pid, SIGTERM)
        except OSError as err:
            err = str(err)
            if err.find('No such process') > 0:
                if os.path.exists(self.pidfile):
                    os.remove(self.pidfile)
                log.info('OK')
            else:
                log.error('FAIL')
                log.error(err)
                exit(1)


class Status(object):
    def __init__(self, pidfile):
        self.pidfile = pidfile

    def execute(self):
        try:
            with file(self.pidfile, 'r') as fh:
                pid = int(fh.read().strip())
        except (IOError, ValueError):
            pid = None

        if not pid:
            log.error('Stopped')
            return

        try:
            os.kill(pid, 0)
            log.info('Running: %d', pid)
        except OSError as err:
            log.error('Pid file exists, but cannot send signal to it - %s', err)



try:
    import httplib
except ImportError:
    import http.client as httplib  # pylint: disable=F0401

from datetime import timedelta
from random import shuffle

import msgpack

import json

from tornado import gen
from tornado import httputil
from tornado.httpclient import AsyncHTTPClient
from tornado.httpclient import HTTPError
from tornado.httpclient import HTTPRequest

from cocaine.exceptions import ServiceError

from cocaine.services import Service, Locator

from cocaine.proxy.helpers import extract_app_and_event
from cocaine.proxy.helpers import fill_response_in
from cocaine.proxy.helpers import pack_httprequest

from cocaine.proxy.plugin import IPlugin
from cocaine.proxy.plugin import PluginApplicationError
from cocaine.proxy.plugin import PluginConfigurationError
from cocaine.proxy.plugin import PluginNoSuchApplication

from cocaine.proxy.proxy import RESOLVE_TIMEOUT, LOCATORCATEGORY, ESERVICENOTAVAILABLE


def is_mds_stid(stid):
    parts = stid.split(".", 2)
    return len(parts) == 3 and parts[2].startswith('E') and ':' in parts[2]


def is_mds_key(key):
    parts = key.split("/", 1)
    return len(parts) == 2 and parts[0].isdigit()


class MDSDirect(IPlugin):
    def __init__(self, proxy, config):
        super(MDSDirect, self).__init__(proxy)
        try:
            self.dist_info_endpoint = config["dist_info_endpoint"]
            self.mds_dist_info_endpoint = config["mds_dist_info_endpoint"]
            self.locator_port = config["locator_port"]
            self.filter_mds_stid = config.get("filter_stid", True)
            self.service_connect_timeout = timedelta(milliseconds=config.get("service_connect_timeout_ms", 1500))
            self.srw_httpclient = AsyncHTTPClient()
        except KeyError as err:
            raise PluginConfigurationError(self.name(), "option required %s" % err)

    @staticmethod
    def name():
        return "mds-direct"

    def match(self, request):
        if "X-Srw-Key" in request.headers and "X-Srw-Key-Type" in request.headers and "X-Srw-Namespace" in request.headers:
            key = request.headers["X-Srw-Key"]
            return not self.filter_mds_stid or is_mds_stid(key) or is_mds_key(key)
        return False

    @gen.coroutine
    def reelect_app(self, request, app):
        """tries to connect to the same app on differnet host from dist-info"""

        # disconnect app explicitly to break possibly existing connection
        app.disconnect()
        endpoints_size = len(app.locator.endpoints)

        # try x times, where x is the number of different endpoints in app locator.
        for _ in xrange(0, endpoints_size + 1):
            # last chance to take app from common pool
            if len(app.locator.endpoints) == 0:
                request.logger.info(
                    "giving up on connecting to dist-info hosts, falling back to common pool processing")
                app = yield self.proxy.reelect_app(request, app)
                raise gen.Return(app)

            try:
                # always create new locator to prevent locking as we do connect with timeout
                # however lock can be still held during TCP timeout
                locator = Locator(endpoints=app.locator.endpoints)
                request.logger.info("connecting to locator %s", locator.endpoints[0])

                # first try to connect to locator only on remote host with timeout
                yield gen.with_timeout(self.service_connect_timeout, locator.connect())
                request.logger.debug("connected to locator %s for %s", locator.endpoints[0], app.name)
                app = Service(app.name, locator=locator, timeout=RESOLVE_TIMEOUT)

                # try to resolve and connect to application itself
                yield gen.with_timeout(self.service_connect_timeout, app.connect())
                request.logger.debug("connected to application %s via %s", app.name, app.endpoints)
            except gen.TimeoutError:
                # on timeout try next endpoint first
                request.logger.warning("timed out while connecting to application")
                continue
            except ServiceError as err:
                request.logger.warning("got error while resolving app - %s", err)
                if err.category in LOCATORCATEGORY and err.code == ESERVICENOTAVAILABLE:
                    # if the application is down - also try next endpoint
                    continue
                else:
                    raise err
            finally:
                # drop first endpoint to start next connection from different endpoint
                # we do this, as default logic of connection attempts in locator do not fit here
                app.locator.endpoints = app.locator.endpoints[1:]
            # return connected app
            raise gen.Return(app)
        raise PluginApplicationError(42, 42, "could not connect to application")

    def is_stid_request(self, request):
        return request.headers["X-Srw-Key-Type"].upper() == "STID"

    @gen.coroutine
    def process(self, request):
        mds_request_headers = httputil.HTTPHeaders()
        if "Authorization" in request.headers:
            mds_request_headers["Authorization"] = request.headers["Authorization"]

        traceid = getattr(request, "traceid", None)
        if traceid is not None:
            mds_request_headers["X-Request-Id"] = traceid

        key = request.headers["X-Srw-Key"]

        name, event = extract_app_and_event(request)
        timeout = self.proxy.get_timeout(name, event)
        name = self.proxy.resolve_group_to_version(name)
        if self.is_stid_request(request):
            url = "%s/gate/dist-info/%s?primary-only" % (self.dist_info_endpoint, key)
            request.logger.debug("fetching endpoints via mulcagate dist-info - %s", url)
            srw_request = HTTPRequest(
                url,
                method="GET",
                headers=mds_request_headers,
                allow_ipv6=True,
                request_timeout=timeout)
        else:
            url = "%s/dist-info-%s/%s" % (self.mds_dist_info_endpoint, request.headers["X-Srw-Namespace"], key)
            request.logger.debug("fetching endpoints via mds dist-info - %s", url)
            srw_request = HTTPRequest(
                url,
                method="GET",
                headers=mds_request_headers,
                allow_ipv6=True,
                request_timeout=timeout)

        endpoints = yield self.fetch_mds_endpoints(request, srw_request)
        locator = Locator(endpoints=endpoints)
        app = Service(name, locator=locator, timeout=RESOLVE_TIMEOUT)
        request.logger.info("connecting to app %s", name)
        app = yield self.reelect_app(request, app)
        # TODO: attempts should be configurable
        yield self.proxy.process(request, name, app, event, pack_httprequest(request), self.reelect_app, 4, timeout)

    def decode_mulca_dist_info(self, body):
        lines = body.split("\n")
        endpoints = [(line.split()[0], self.locator_port) for line in lines if line]
        shuffle(endpoints)
        return endpoints

    def decode_mds_dist_info(self, body):
        obj = json.loads(body)
        endpoints = [(x['host'], self.locator_port) for x in obj['primary']]
        shuffle(endpoints)
        return endpoints

    @gen.coroutine
    def fetch_mds_endpoints(self, request, srw_request):
        try:
            # NOTE: we can do it in a streaming way
            resp = yield self.srw_httpclient.fetch(srw_request)
            body = resp.buffer.read(None)
            if self.is_stid_request(request):
                raise gen.Return(self.decode_mulca_dist_info(body))
            else:
                raise gen.Return(self.decode_mds_dist_info(body))

        except HTTPError as err:
            if err.code == 404:
                raise PluginNoSuchApplication("404")

            if err.code == 500:
                raise PluginApplicationError(42, 42, "500")

            if err.code == 401:
                fill_response_in(request, err.code,
                                 httplib.responses.get(err.code, httplib.OK),
                                 err.response.body, err.response.headers)
                return

            raise err


def decode_chunked_encoded_reply(resp):
    # read_size is set to to prevent overead from BytesIO
    # in this case the rest of the buffer is not packed data
    code, raw_headers = msgpack.Unpacker(resp.buffer, read_size=1).unpack()
    body = resp.buffer.read(None)
    headers = httputil.HTTPHeaders(raw_headers)
    return code, headers, body



import msgpack
from tornado import gen

from cocaine.decorators import coroutine
from cocaine.exceptions import ServiceError
from cocaine.tools import log
from cocaine.tools.actions import Action
from cocaine.tools.error import ToolsError

_COLLECTION = '.collection-acls'
_TAGS = ('storage-acls',)


class List(Action):
    def __init__(self, storage):
        self._storage = storage

    @coroutine
    def execute(self):
        channel = yield self._storage.find(_COLLECTION, _TAGS)
        result = yield channel.rx.get()
        raise gen.Return(result)


class View(Action):
    def __init__(self, name, storage):
        self._name = name
        self._storage = storage

    @coroutine
    def execute(self):
        channel = yield self._storage.read(_COLLECTION, self._name)
        result = yield channel.rx.get()
        raise gen.Return(msgpack.loads(result))


class Create(Action):
    def __init__(self, name, storage):
        self._name = name
        self._storage = storage

    @coroutine
    def execute(self):
        try:
            yield View(self._name, self._storage).execute()
        except ServiceError:
            pass
        else:
            raise ToolsError('An ACL for collection "{}" already exists'.format(self._name))

        channel = yield self._storage.write(_COLLECTION, self._name, msgpack.dumps([{}, {}]), _TAGS)
        yield channel.rx.get()


class Edit(Action):
    PERM = {
        '0': 0b00,
        'R': 0b01,
        'W': 0b10,
        'RW': 0b11,
    }

    def __init__(self, name, cids, uids, perm, storage):
        self._name = name
        self._cids = cids
        self._uids = uids
        self._perm = self.PERM.get(perm, 0)
        self._storage = storage

    @coroutine
    def execute(self):
        try:
            content = yield View(self._name, self._storage).execute()
        except ServiceError:
            content = [{}, {}]

        if len(content) != 2:
            raise ToolsError('framing error - ACL should be a tuple of 2 maps')

        cids, uids = content
        for cid in self._cids:
            cids[int(cid)] = self._perm
        for uid in self._uids:
            uids[int(uid)] = self._perm

        content = msgpack.dumps([cids, uids])
        channel = yield self._storage.write(_COLLECTION, self._name, content, _TAGS)
        yield channel.rx.get()


class RemoveAclError(ToolsError):
    def __init__(self, failed):
        self._failed = failed

    @property
    def failed(self):
        return self._failed

    def __str__(self):
        reasons = []
        for acl, err in self._failed:
            reasons.append(' - "{}": {}'.format(acl, err.reason.lower()))
        return 'Failed to remove all ACL:\n{}'.format('\n'.join(reasons))


class Remove(Action):
    def __init__(self, name, storage):
        self._name = name
        self._storage = storage

    @coroutine
    def execute(self):
        if self._name is None:
            acls = yield List(self._storage).execute()
        else:
            acls = [self._name]

        log.info('ACL to be removed: %s', acls)

        failed = []
        for acl in acls:
            try:
                channel = yield self._storage.remove(_COLLECTION, acl)
                yield channel.rx.get()
            except ServiceError as err:
                failed.append((acl, err))
            else:
                log.info('ACL %s has been successfully removed', acl)

        if len(failed) > 0:
            raise RemoveAclError(failed)
